r"""
Compute invariants of quintics and sextics via 'Ueberschiebung'

.. TODO::

    * Implement invariants in small positive characteristic.

    * Cardona-Quer and additional invariants for classifying automorphism groups.

AUTHORS:

- Nick Alexander
- Sabrina Kunzweiler, Gareth Ma, Giacomo Pope (2024): adapt to smooth model

"""

# ****************************************************************************
#       Copyright (C) 2008 Nick Alexander
#                     2025 Sabrina Kunzweiler <sabrina.kunzweiler@math.u-bordeaux.fr>
#                     2025 Gareth Ma <grhkm21@gmail.com>
#                     2025 Giacomo Pope <giacomopope@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing


def diffxy(f, x, xtimes, y, ytimes):
    r"""
    Differentiate a polynomial ``f``, ``xtimes`` with respect to ``x``, and
    ```ytimes`` with respect to ``y``.

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import diffxy
        sage: R.<u, v> = QQ[]
        sage: diffxy(u^2*v^3, u, 0, v, 0)
        u^2*v^3
        sage: diffxy(u^2*v^3, u, 2, v, 1)
        6*v^2
        sage: diffxy(u^2*v^3, u, 2, v, 2)
        12*v
        sage: diffxy(u^2*v^3 + u^4*v^4, u, 2, v, 2)
        144*u^2*v^2 + 12*v
    """
    h = f
    for i in range(xtimes):
        h = h.derivative(x)
    for j in range(ytimes):
        h = h.derivative(y)
    return h


def differential_operator(f, g, k):
    r"""
    Return the differential operator `(f g)_k` symbolically in the polynomial ring in ``dfdx, dfdy, dgdx, dgdy``.

    This is defined by Mestre on p 315 [Mes1991]_:

    .. MATH::

        (f g)_k = \frac{(m - k)! (n - k)!}{m! n!} \left(
        \frac{\partial f}{\partial x} \frac{\partial g}{\partial y} -
        \frac{\partial f}{\partial y} \frac{\partial g}{\partial x} \right)^k .

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import differential_operator
        sage: R.<x, y> = QQ[]
        sage: differential_operator(x, y, 0)
        1
        sage: differential_operator(x, y, 1)
        -dfdy*dgdx + dfdx*dgdy
        sage: differential_operator(x*y, x*y, 2)
        1/4*dfdy^2*dgdx^2 - 1/2*dfdx*dfdy*dgdx*dgdy + 1/4*dfdx^2*dgdy^2
        sage: differential_operator(x^2*y, x*y^2, 2)
        1/36*dfdy^2*dgdx^2 - 1/18*dfdx*dfdy*dgdx*dgdy + 1/36*dfdx^2*dgdy^2
        sage: differential_operator(x^2*y, x*y^2, 4)
        1/576*dfdy^4*dgdx^4 - 1/144*dfdx*dfdy^3*dgdx^3*dgdy + 1/96*dfdx^2*dfdy^2*dgdx^2*dgdy^2
        - 1/144*dfdx^3*dfdy*dgdx*dgdy^3 + 1/576*dfdx^4*dgdy^4
    """
    (x, y) = f.parent().gens()
    n = max(ZZ(f.degree()), ZZ(k))
    m = max(ZZ(g.degree()), ZZ(k))
    R, (fx, fy, gx, gy) = PolynomialRing(
        f.base_ring(), 4, "dfdx,dfdy,dgdx,dgdy"
    ).objgens()
    const = (m - k).factorial() * (n - k).factorial() / (m.factorial() * n.factorial())
    return f.base_ring()(const) * (fx * gy - fy * gx) ** k


def diffsymb(U, f, g):
    r"""
    Given a differential operator ``U`` in ``dfdx, dfdy, dgdx, dgdy``,
    represented symbolically by ``U``, apply it to ``f, g``.

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import diffsymb
        sage: R.<x, y> = QQ[]
        sage: S.<dfdx, dfdy, dgdx, dgdy> = QQ[]
        sage: [ diffsymb(dd, x^2, y*0 + 1) for dd in S.gens() ]
        [2*x, 0, 0, 0]
        sage: [ diffsymb(dd, x*0 + 1, y^2) for dd in S.gens() ]
        [0, 0, 0, 2*y]
        sage: [ diffsymb(dd, x^2, y^2) for dd in S.gens() ]
        [2*x*y^2, 0, 0, 2*x^2*y]

        sage: diffsymb(dfdx + dfdy*dgdy, y*x^2, y^3)
        2*x*y^4 + 3*x^2*y^2
    """
    (x, y) = f.parent().gens()
    R, (fx, fy, gx, gy) = PolynomialRing(
        f.base_ring(), 4, "dfdx,dfdy,dgdx,dgdy"
    ).objgens()
    res = 0
    for coeff, mon in list(U):
        m = R(mon)
        a = diffxy(f, x, m.degree(fx), y, m.degree(fy))
        b = diffxy(g, x, m.degree(gx), y, m.degree(gy))
        temp = coeff * a * b
        res = res + temp
    return res


def Ueberschiebung(f, g, k):
    r"""
    Return the differential operator `(f g)_k`.

    This is defined by Mestre on page 315 [Mes1991]_:

    .. MATH::

        (f g)_k = \frac{(m - k)! (n - k)!}{m! n!} \left(
        \frac{\partial f}{\partial x} \frac{\partial g}{\partial y} -
        \frac{\partial f}{\partial y} \frac{\partial g}{\partial x} \right)^k .

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import Ueberschiebung as ub
        sage: R.<x, y> = QQ[]
        sage: ub(x, y, 0)
        x*y
        sage: ub(x^5 + 1, x^5 + 1, 1)
        0
        sage: ub(x^5 + 5*x + 1, x^5 + 5*x + 1, 0)
        x^10 + 10*x^6 + 2*x^5 + 25*x^2 + 10*x + 1
    """
    U = differential_operator(f, g, k)
    # U is the (f g)_k = ... of Mestre, p315, symbolically
    return diffsymb(U, f, g)


def ubs(f):
    r"""
    Given a sextic form `f`, return a dictionary of the invariants of Mestre, p 317 [Mes1991]_.

    `f` may be homogeneous in two variables or inhomogeneous in one.

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import ubs
        sage: x = QQ['x'].0
        sage: ubs(x^6 + 1)
        {'A': 2,
         'B': 2/3,
         'C': -2/9,
         'D': 0,
         'Delta': -2/3*x^2*h^2,
         'f': x^6 + h^6,
         'i': 2*x^2*h^2,
         'y1': 0,
         'y2': 0,
         'y3': 0}

        sage: R.<u, v> = QQ[]
        sage: ubs(u^6 + v^6)
        {'A': 2,
         'B': 2/3,
         'C': -2/9,
         'D': 0,
         'Delta': -2/3*u^2*v^2,
         'f': u^6 + v^6,
         'i': 2*u^2*v^2,
         'y1': 0,
         'y2': 0,
         'y3': 0}

        sage: R.<t> = GF(31)[]
        sage: ubs(t^6 + 2*t^5 + t^2 + 3*t + 1)
        {'A': 0,
         'B': -12,
         'C': -15,
         'D': -15,
         'Delta': -10*t^4 + 12*t^3*h + 7*t^2*h^2 - 5*t*h^3 + 2*h^4,
         'f': t^6 + 2*t^5*h + t^2*h^4 + 3*t*h^5 + h^6,
         'i': -4*t^4 + 10*t^3*h + 2*t^2*h^2 - 9*t*h^3 - 7*h^4,
         'y1': 4*t^2 - 10*t*h - 13*h^2,
         'y2': 6*t^2 - 4*t*h + 2*h^2,
         'y3': 4*t^2 - 4*t*h - 9*h^2}
    """
    ub = Ueberschiebung
    if f.parent().ngens() == 1:
        f = PolynomialRing(f.parent().base_ring(), 1, f.parent().variable_name())(f)
        x1, x2 = f.homogenize().parent().gens()
        f = sum([f[i] * x1**i * x2 ** (6 - i) for i in range(7)])
    U = {}
    U["f"] = f
    U["i"] = ub(f, f, 4)
    U["Delta"] = ub(U["i"], U["i"], 2)
    U["y1"] = ub(f, U["i"], 4)
    U["y2"] = ub(U["i"], U["y1"], 2)
    U["y3"] = ub(U["i"], U["y2"], 2)
    U["A"] = ub(f, f, 6)
    U["B"] = ub(U["i"], U["i"], 4)
    U["C"] = ub(U["i"], U["Delta"], 4)
    U["D"] = ub(U["y3"], U["y1"], 2)
    return U


def clebsch_to_igusa(A, B, C, D):
    r"""
    Convert Clebsch invariants `A, B, C, D` to Igusa invariants `I_2, I_4, I_6, I_{10}`.

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import clebsch_to_igusa, igusa_to_clebsch
        sage: clebsch_to_igusa(2, 3, 4, 5)
        (-240, 17370, 231120, -103098906)
        sage: igusa_to_clebsch(*clebsch_to_igusa(2, 3, 4, 5))
        (2, 3, 4, 5)

        sage: Cs = tuple(map(GF(31), (2, 3, 4, 5))); Cs
        (2, 3, 4, 5)
        sage: clebsch_to_igusa(*Cs)
        (8, 10, 15, 26)
        sage: igusa_to_clebsch(*clebsch_to_igusa(*Cs))
        (2, 3, 4, 5)
    """
    I2 = -120 * A
    I4 = -720 * A**2 + 6750 * B
    I6 = 8640 * A**3 - 108000 * A * B + 202500 * C
    I10 = (
        -62208 * A**5
        + 972000 * A**3 * B
        + 1620000 * A**2 * C
        - 3037500 * A * B**2
        - 6075000 * B * C
        - 4556250 * D
    )
    return (I2, I4, I6, I10)


def igusa_to_clebsch(I2, I4, I6, I10):
    r"""
    Convert Igusa invariants `I_2, I_4, I_6, I_{10}` to Clebsch invariants `A, B, C, D`.

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import clebsch_to_igusa, igusa_to_clebsch
        sage: igusa_to_clebsch(-2400, 173700, 23112000, -10309890600)
        (20, 342/5, 2512/5, 43381012/1125)
        sage: clebsch_to_igusa(*igusa_to_clebsch(-2400, 173700, 23112000, -10309890600))
        (-2400, 173700, 23112000, -10309890600)

        sage: Is = tuple(map(GF(31), (-2400, 173700, 23112000, -10309890600))); Is
        (18, 7, 12, 27)
        sage: igusa_to_clebsch(*Is)
        (20, 25, 25, 12)
        sage: clebsch_to_igusa(*igusa_to_clebsch(*Is))
        (18, 7, 12, 27)
    """
    A = -(+I2) / 120
    B = -(-(I2**2) - 20 * I4) / 135000
    C = -(+(I2**3) + 80 * I2 * I4 - 600 * I6) / 121500000
    D = (
        -(
            +9 * I2**5
            + 700 * I2**3 * I4
            - 3600 * I2**2 * I6
            - 12400 * I2 * I4**2
            + 48000 * I4 * I6
            + 10800000 * I10
        )
        / 49207500000000
    )
    return (A, B, C, D)


def clebsch_invariants(f):
    r"""
    Given a sextic form `f`, return the Clebsch invariants `(A, B, C, D)` of
    Mestre, p 317, [Mes1991]_.

    `f` may be homogeneous in two variables or inhomogeneous in one.

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import clebsch_invariants
        sage: R.<x, y> = QQ[]
        sage: clebsch_invariants(x^6 + y^6)
        (2, 2/3, -2/9, 0)
        sage: R.<x> = QQ[]
        sage: clebsch_invariants(x^6 + x^5 + x^4 + x^2 + 2)
        (62/15, 15434/5625, -236951/140625, 229930748/791015625)

        sage: magma(x^6 + 1).ClebschInvariants()                    # optional - magma
        [ 2, 2/3, -2/9, 0 ]
        sage: magma(x^6 + x^5 + x^4 + x^2 + 2).ClebschInvariants()  # optional - magma
        [ 62/15, 15434/5625, -236951/140625, 229930748/791015625 ]
    """
    R = f.parent().base_ring()
    if R.characteristic() in [2, 3, 5]:
        raise NotImplementedError(
            "Invariants of binary sextics/genus 2 hyperelliptic "
            "curves not implemented in characteristics 2, 3, and 5"
        )

    U = ubs(f)
    L = U["A"], U["B"], U["C"], U["D"]
    assert all(t.is_constant() for t in L)
    return tuple([t.constant_coefficient() for t in L])


def igusa_clebsch_invariants(f):
    r"""
    Given a sextic form `f`, return the Igusa-Clebsch invariants `I_2, I_4,
    I_6, I_{10}` of Igusa and Clebsch [IJ1960]_.

    `f` may be homogeneous in two variables or inhomogeneous in one.

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import igusa_clebsch_invariants
        sage: R.<x, y> = QQ[]
        sage: igusa_clebsch_invariants(x^6 + y^6)
        (-240, 1620, -119880, -46656)
        sage: R.<x> = QQ[]
        sage: igusa_clebsch_invariants(x^6 + x^5 + x^4 + x^2 + 2)
        (-496, 6220, -955932, -1111784)

        sage: magma(x^6 + 1).IgusaClebschInvariants()                    # optional - magma
        [ -240, 1620, -119880, -46656 ]
        sage: magma(x^6 + x^5 + x^4 + x^2 + 2).IgusaClebschInvariants()  # optional - magma
        [ -496, 6220, -955932, -1111784 ]

    TESTS:

    Let's check a symbolic example::

        sage: R.<a, b, c, d, e> = QQ[]
        sage: S.<x> = R[]
        sage: igusa_clebsch_invariants(x^5 + a*x^4 + b*x^3 + c*x^2 + d*x + e)[0]
        6*b^2 - 16*a*c + 40*d

        sage: from sage.schemes.hyperelliptic_curves.invariants import absolute_igusa_invariants_wamelen
        sage: absolute_igusa_invariants_wamelen(GF(5)['x'](x^6 - 2*x))
        Traceback (most recent call last):
        ...
        NotImplementedError: Invariants of binary sextics/genus 2 hyperelliptic curves
        not implemented in characteristics 2, 3, and 5
    """
    return clebsch_to_igusa(*clebsch_invariants(f))


def absolute_igusa_invariants_wamelen(f):
    r"""
    Given a sextic form `f`, return the three absolute Igusa invariants used by van Wamelen [Wam1999]_.

    `f` may be homogeneous in two variables or inhomogeneous in one.

    REFERENCES:

    - [Wam1999]_

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import absolute_igusa_invariants_wamelen
        sage: R.<x> = QQ[]
        sage: absolute_igusa_invariants_wamelen(x^5 - 1)
        (0, 0, 0)

    The following example can be checked against van Wamelen's paper::

        sage: h = -x^5 + 3*x^4 + 2*x^3 - 6*x^2 - 3*x + 1
        sage: i1, i2, i3 = absolute_igusa_invariants_wamelen(h)
        sage: list(map(factor, (i1, i2, i3)))
        [2^7 * 3^15, 2^5 * 3^11 * 5, 2^4 * 3^9 * 31]

    TESTS::

        sage: absolute_igusa_invariants_wamelen(GF(3)['x'](x^5 - 2*x))
        Traceback (most recent call last):
        ...
        NotImplementedError: Invariants of binary sextics/genus 2 hyperelliptic curves
        not implemented in characteristics 2, 3, and 5
    """
    I2, I4, I6, I10 = igusa_clebsch_invariants(f)
    i1 = I2**5 / I10
    i2 = I2**3 * I4 / I10
    i3 = I2**2 * I6 / I10
    return (i1, i2, i3)


def absolute_igusa_invariants_kohel(f):
    r"""
    Given a sextic form `f`, return the three absolute Igusa invariants used by Kohel [KohECHIDNA]_.

    `f` may be homogeneous in two variables or inhomogeneous in one.

    EXAMPLES::

        sage: from sage.schemes.hyperelliptic_curves.invariants import absolute_igusa_invariants_kohel
        sage: R.<x> = QQ[]
        sage: absolute_igusa_invariants_kohel(x^5 - 1)
        (0, 0, 0)
        sage: absolute_igusa_invariants_kohel(x^5 - x)
        (100, -20000, -2000)

    The following example can be checked against Kohel's database [KohECHIDNA]_ ::

        sage: h = -x^5 + 3*x^4 + 2*x^3 - 6*x^2 - 3*x + 1
        sage: i1, i2, i3 = absolute_igusa_invariants_kohel(h)
        sage: list(map(factor, (i1, i2, i3)))
        [2^2 * 3^5 * 5 * 31, 2^5 * 3^11 * 5, 2^4 * 3^9 * 31]
        sage: list(map(factor, (150660, 28343520, 9762768)))
        [2^2 * 3^5 * 5 * 31, 2^5 * 3^11 * 5, 2^4 * 3^9 * 31]

    TESTS::

        sage: absolute_igusa_invariants_kohel(GF(2)['x'](x^5 - x))
        Traceback (most recent call last):
        ...
        NotImplementedError: Invariants of binary sextics/genus 2 hyperelliptic curves
        not implemented in characteristics 2, 3, and 5
    """
    I2, I4, I6, I10 = igusa_clebsch_invariants(f)
    i1 = I4 * I6 / I10
    i2 = I2**3 * I4 / I10
    i3 = I2**2 * I6 / I10
    return (i1, i2, i3)



# distutils: libraries = glpk z gmp

cdef extern from "glpk.h":
    int glp_init_env()
    int glp_free_env()
    int glp_term_out(int flag)
    void glp_term_hook(int (*func)(void *info, const char *s), void *info)
    int glp_open_tee(const char *fname)
    int glp_close_tee()
    void glp_error_hook(void (*func)(void *info), void *info)
    int glp_at_error()
    void glp_mem_limit(int limit)
    void glp_mem_usage(int *count, int *cpeak, size_t *total, size_t *tpeak)



# sage.doctest: needs sage.libs.flint
"""
Numerical computation of newforms
"""

# ****************************************************************************
#       Copyright (C) 2004-2006 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.fast_arith import prime_range
from sage.matrix.constructor import matrix
from sage.misc.verbose import verbose
from sage.misc.cachefunc import cached_method
from sage.misc.prandom import randint
from sage.modular.arithgroup.congroup_gamma0 import Gamma0_constructor as Gamma0
from sage.modular.modsym.all import ModularSymbols
from sage.modules.free_module_element import free_module_element as vector
from sage.rings.complex_double import CDF
from sage.rings.integer import Integer
from sage.rings.rational_field import QQ
from sage.structure.richcmp import richcmp_method, richcmp
from sage.structure.sage_object import SageObject
from sage.structure.sequence import Sequence

# This variable controls importing the SciPy library sparingly
scipy = None


@richcmp_method
class NumericalEigenforms(SageObject):
    """
    numerical_eigenforms(group, weight=2, eps=1e-20, delta=1e-2, tp=[2,3,5]).

    INPUT:

    - ``group`` -- a congruence subgroup of a Dirichlet character of
      order 1 or 2

    - ``weight`` -- integer >= 2

    - ``eps`` -- a small float; abs( ) < eps is what "equal to zero" is
      interpreted as for floating point numbers

    - ``delta`` -- a small-ish float; eigenvalues are considered distinct
      if their difference has absolute value at least delta

    - ``tp`` -- use the Hecke operators T_p for p in tp when searching
      for a random Hecke operator with distinct Hecke eigenvalues

    OUTPUT: a numerical eigenforms object, with the following useful methods:

    - :meth:`ap` -- return all eigenvalues of `T_p`

    - :meth:`eigenvalues` -- list of eigenvalues corresponding
      to the given list of primes, e.g.,::

          [[eigenvalues of T_2],
           [eigenvalues of T_3],
           [eigenvalues of T_5], ...]

    - :meth:`systems_of_eigenvalues` -- list of the systems of
      eigenvalues of eigenforms such that the chosen random linear
      combination of Hecke operators has multiplicity 1 eigenvalues.

    EXAMPLES::

        sage: n = numerical_eigenforms(23)
        sage: n == loads(dumps(n))
        True
        sage: n.ap(2)  # abs tol 1e-11
        [3.0, -1.6180339887498947, 0.6180339887498968]
        sage: n.systems_of_eigenvalues(7)  # abs tol 1e-11
        [[-1.6180339887498947, 2.23606797749979, -3.2360679774997894],
         [0.618033988749895, -2.236067977499788, 1.23606797749979],
         [3.0, 4.0, 6.0]]
        sage: n.systems_of_abs(7)  # abs tol 1e-11
        [[0.618033988749895, 2.236067977499788, 1.23606797749979],
         [1.6180339887498947, 2.23606797749979, 3.2360679774997894],
         [3.0, 4.0, 6.0]]
        sage: n.eigenvalues([2,3,5])  # abs tol 1e-11
        [[3.0, -1.6180339887498947, 0.6180339887498968],
         [4.0, 2.2360679774997894, -2.236067977499788],
         [6.0, -3.2360679774997894, 1.2360679774997936]]
    """
    def __init__(self, group, weight=2, eps=1e-20,
                 delta=1e-2, tp=[2, 3, 5]):
        """
        Create a new space of numerical eigenforms.

        EXAMPLES::

            sage: numerical_eigenforms(61) # indirect doctest
            Numerical Hecke eigenvalues for Congruence Subgroup Gamma0(61) of weight 2
        """
        if isinstance(group, (int, Integer)):
            group = Gamma0(Integer(group))
        self._group = group
        self._weight = Integer(weight)
        self._tp = tp
        if self._weight < 2:
            raise ValueError("weight must be at least 2")
        self._eps = eps
        self._delta = delta

    def __richcmp__(self, other, op):
        """
        Compare two spaces of numerical eigenforms.

        They are considered equal if and only if they come from the
        same space of modular symbols.

        EXAMPLES::

            sage: n = numerical_eigenforms(23)
            sage: n == loads(dumps(n))
            True
        """
        if not isinstance(other, NumericalEigenforms):
            return NotImplemented
        return richcmp(self.modular_symbols(), other.modular_symbols(), op)

    def level(self):
        """
        Return the level of this set of modular eigenforms.

        EXAMPLES::

            sage: n = numerical_eigenforms(61) ; n.level()
            61
        """
        return self._group.level()

    def weight(self):
        """
        Return the weight of this set of modular eigenforms.

        EXAMPLES::

            sage: n = numerical_eigenforms(61) ; n.weight()
            2
        """
        return self._weight

    def _repr_(self):
        """
        Print string representation of ``self``.

        EXAMPLES::

            sage: n = numerical_eigenforms(61) ; n
            Numerical Hecke eigenvalues for Congruence Subgroup Gamma0(61) of weight 2

            sage: n._repr_()
            'Numerical Hecke eigenvalues for Congruence Subgroup Gamma0(61) of weight 2'
        """
        return "Numerical Hecke eigenvalues for %s of weight %s" % (
            self._group, self._weight)

    @cached_method
    def modular_symbols(self):
        """
        Return the space of modular symbols used for computing this
        set of modular eigenforms.

        EXAMPLES::

            sage: n = numerical_eigenforms(61) ; n.modular_symbols()
            Modular Symbols space of dimension 5 for Gamma_0(61) of weight 2 with sign 1 over Rational Field
        """
        M = ModularSymbols(self._group,
                self._weight, sign=1)
        if M.base_ring() != QQ:
            raise ValueError("modular forms space must be defined over QQ")
        return M

    @cached_method
    def _eigenvectors(self):
        r"""
        Find numerical approximations to simultaneous eigenvectors in
        self.modular_symbols() for all T_p in self._tp.

        EXAMPLES::

            sage: n = numerical_eigenforms(61)
            sage: n._eigenvectors() # random order
            [              1.0    0.289473640239    0.176788851952    0.336707726757  2.4182243084e-16]
            [                0  -0.0702748344418    0.491416161212    0.155925712173    0.707106781187]
            [                0    0.413171180356    0.141163094698   0.0923242547901    0.707106781187]
            [                0    0.826342360711    0.282326189397     0.18464850958 6.79812569682e-16]
            [                0      0.2402380858    0.792225196393    0.905370774276 4.70805946682e-16]

        TESTS:

        This tests if this routine selects only eigenvectors with
        multiplicity one.  Two of the eigenvalues are
        (roughly) -92.21 and -90.30 so if we set ``eps = 2.0``
        then they should compare as equal, causing both eigenvectors
        to be absent from the matrix returned.  The remaining eigenvalues
        (ostensibly unique) are visible in the test, which should be
        independent of which eigenvectors are returned, but it does presume
        an ordering of these eigenvectors for the test to succeed.
        This exercises a correction in :issue:`8018`. ::

            sage: n = numerical_eigenforms(61, eps=2.0)
            sage: evectors = n._eigenvectors()
            sage: evalues = [(matrix((n._hecke_matrix*evectors).column(i))/matrix(evectors.column(i)))[0, 0]
            ....:            for i in range(evectors.ncols())]
            sage: diff = n._hecke_matrix*evectors - evectors*diagonal_matrix(evalues)
            sage: sum(abs(a) for a in diff.list()) < 1.0e-9
            True
        """
        verbose('Finding eigenvector basis')
        M = self.modular_symbols()

        tp = self._tp
        p = tp[0]
        t = M.T(p).matrix()
        for p in tp[1:]:
            t += randint(-50, 50) * M.T(p).matrix()

        self._hecke_matrix = t

        global scipy
        if scipy is None:
            import scipy
        import scipy.linalg
        evals, eig = scipy.linalg.eig(self._hecke_matrix.numpy(), right=True, left=False)
        B = matrix(eig)
        v = [CDF(evals[i]) for i in range(len(evals))]

        # Determine the eigenvectors with eigenvalues of multiplicity
        # one, with equality controlled by the value of eps
        # Keep just these eigenvectors
        eps = self._eps
        w = []
        for i in range(len(v)):
            e = v[i]
            uniq = True
            for j in range(len(v)):
                if uniq and i != j and abs(e-v[j]) < eps:
                    uniq = False
            if uniq:
                w.append(i)
        return B.matrix_from_columns(w)

    @cached_method
    def _easy_vector(self):
        """
        Return a very sparse vector v such that v times the eigenvector matrix
        has all entries nonzero.

        ALGORITHM:

        1. Choose row with the most nonzero entries.   (put 1 there)

        2. Consider submatrix of columns corresponding to zero entries
           in row chosen in 1.

        3. Find row of submatrix with most nonzero entries, and add
           appropriate multiple.  Repeat.

        EXAMPLES::

            sage: n = numerical_eigenforms(37)
            sage: n._easy_vector()                 # slightly random output
            (1.0, 1.0, 0)
            sage: n = numerical_eigenforms(43)
            sage: n._easy_vector()                 # slightly random output
            (1.0, 0, 1.0, 0)
            sage: n = numerical_eigenforms(125)
            sage: n._easy_vector()                 # slightly random output
            (0, 0, 0, 1.0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        """
        E = self._eigenvectors()
        delta = self._delta
        x = (CDF**E.nrows()).zero_vector()
        if E.nrows() == 0:
            return x

        def best_row(M):
            """
            Find the best row among rows of M, i.e. the row
            with the most entries supported outside [-delta, delta].

            EXAMPLES::

                sage: numerical_eigenforms(61)._easy_vector() # indirect doctest
                (1.0, 0.0, 0.0, 0.0, 1.0)
            """
            R = M.rows()
            v = [len(support(r, delta)) for r in R]
            m = max(v)
            i = v.index(m)
            return i, R[i]

        i, e = best_row(E)

        x[i] = 1

        while True:
            s = set(support(e, delta))
            zp = [j for j in range(e.degree()) if j not in s]
            if not zp:
                break
            C = E.matrix_from_columns(zp)
            # best row
            i, f = best_row(C)
            x[i] += 1   # simplistic
            e = x * E

        self.__easy_vector = x
        return x

    @cached_method
    def _eigendata(self):
        """
        Return all eigendata for self._easy_vector().

        EXAMPLES::

            sage: numerical_eigenforms(61)._eigendata() # random order
            ((1.0, 0.668205013164, 0.219198805797, 0.49263343893, 0.707106781187), (1.0, 1.49654668896, 4.5620686498, 2.02990686579, 1.41421356237), [0, 1], (1.0, 1.0))
        """
        x = self._easy_vector()

        B = self._eigenvectors()

        def phi(y):
            """
            Take coefficients and a basis, and return that
            linear combination of basis vectors.

            EXAMPLES::

                sage: n = numerical_eigenforms(61) # indirect doctest
                sage: n._eigendata() # random order
                ((1.0, 0.668205013164, 0.219198805797, 0.49263343893, 0.707106781187), (1.0, 1.49654668896, 4.5620686498, 2.02990686579, 1.41421356237), [0, 1], (1.0, 1.0))
            """
            return y.element() * B

        phi_x = phi(x)
        V = phi_x.parent()
        phi_x_inv = V([a**(-1) for a in phi_x])
        eps = self._eps
        nzp = support(x, eps)
        x_nzp = vector(CDF, x.list_from_positions(nzp))
        self.__eigendata = (phi_x, phi_x_inv, nzp, x_nzp)
        return self.__eigendata

    @cached_method
    def ap(self, p):
        """
        Return a list of the eigenvalues of the Hecke operator `T_p`
        on all the computed eigenforms.  The eigenvalues match up
        between one prime and the next.

        INPUT:

        - ``p`` -- integer; a prime number

        OUTPUT: list of double precision complex numbers

        EXAMPLES::

            sage: n = numerical_eigenforms(11,4)
            sage: n.ap(2) # random order
            [9.0, 9.0, 2.73205080757, -0.732050807569]
            sage: n.ap(3) # random order
            [28.0, 28.0, -7.92820323028, 5.92820323028]
            sage: m = n.modular_symbols()
            sage: x = polygen(QQ, 'x')
            sage: m.T(2).charpoly('x').factor()
            (x - 9)^2 * (x^2 - 2*x - 2)
            sage: m.T(3).charpoly('x').factor()
            (x - 28)^2 * (x^2 + 2*x - 47)
        """
        p = Integer(p)
        if not p.is_prime():
            raise ValueError("p must be a prime")
        return Sequence(self.eigenvalues([p])[0], immutable=True)

    def eigenvalues(self, primes):
        """
        Return the eigenvalues of the Hecke operators corresponding
        to the primes in the input list of primes.   The eigenvalues
        match up between one prime and the next.

        INPUT:

        - ``primes`` -- list of primes

        OUTPUT: list of lists of eigenvalues

        EXAMPLES::

            sage: n = numerical_eigenforms(1,12)
            sage: n.eigenvalues([3,5,13])  # rel tol 2.4e-10
            [[177148.0, 252.00000000001896], [48828126.0, 4830.000000001376], [1792160394038.0, -577737.9999898539]]
        """
        primes = [Integer(p) for p in primes]
        for p in primes:
            if not p.is_prime():
                raise ValueError('each element of primes must be prime.')
        phi_x, phi_x_inv, nzp, x_nzp = self._eigendata()
        B = self._eigenvectors()

        def phi(y):
            """
            Take coefficients and a basis, and return that
            linear combination of basis vectors.

            EXAMPLES::

                sage: n = numerical_eigenforms(1,12)  # indirect doctest
                sage: n.eigenvalues([3,5,13])  # rel tol 2.4e-10
                [[177148.0, 252.00000000001896], [48828126.0, 4830.000000001376], [1792160394038.0, -577737.9999898539]]
            """
            return y.element() * B

        ans = []
        m = self.modular_symbols().ambient_module()
        for p in primes:
            t = m._compute_hecke_matrix_prime(p, nzp)
            w = phi(x_nzp*t)
            ans.append([w[i]*phi_x_inv[i] for i in range(w.degree())])
        return ans

    def systems_of_eigenvalues(self, bound):
        """
        Return all systems of eigenvalues for ``self`` for primes
        up to bound.

        EXAMPLES::

            sage: numerical_eigenforms(61).systems_of_eigenvalues(10)  # rel tol 1e-9
            [[-1.481194304092014,
              0.8060634335253706,
              3.156325174658664,
              0.6751308705666462],
             [-1.0, -2.0, -3.0, 1.0],
             [0.311107817465981,
              2.903211925911551,
              -2.5254275608435184,
              -3.214319743377534],
             [2.1700864866260323,
              -1.7092753594369237,
              -1.6308976138151459,
              -0.460811127189112],
             [3.0, 4.0, 6.0, 8.0]]
        """
        P = prime_range(bound)
        e = self.eigenvalues(P)
        v = Sequence([], cr=True)
        if len(e) == 0:
            return v
        for i in range(len(e[0])):
            v.append([e[j][i] for j in range(len(e))])
        v.sort()
        v.set_immutable()
        return v

    def systems_of_abs(self, bound):
        """
        Return the absolute values of all systems of eigenvalues for
        ``self`` for primes up to bound.

        EXAMPLES::

            sage: numerical_eigenforms(61).systems_of_abs(10)  # rel tol 1e-9
            [[0.311107817465981, 2.903211925911551, 2.5254275608435184, 3.214319743377534],
             [1.0, 2.0, 3.0, 1.0],
             [1.481194304092014,
              0.8060634335253706,
              3.156325174658664,
              0.6751308705666462],
             [2.1700864866260323,
              1.7092753594369237,
              1.6308976138151459,
              0.460811127189112],
             [3.0, 4.0, 6.0, 8.0]]
        """
        P = prime_range(bound)
        e = self.eigenvalues(P)
        v = Sequence([], cr=True)
        if len(e) == 0:
            return v
        for i in range(len(e[0])):
            v.append([abs(e[j][i]) for j in range(len(e))])
        v.sort()
        v.set_immutable()
        return v


def support(v, eps):
    """
    Given a vector `v` and a threshold eps, return all
    indices where `|v|` is larger than eps.

    EXAMPLES::

        sage: sage.modular.modform.numerical.support( numerical_eigenforms(61)._easy_vector(), 1.0 )
        []

        sage: sage.modular.modform.numerical.support( numerical_eigenforms(61)._easy_vector(), 0.5 )
        [0, 4]
    """
    return [i for i in range(v.degree()) if abs(v[i]) > eps]



# sage.doctest: needs sage.modules sage.rings.finite_rings
r"""
Hamming codes

Given an integer `r` and a field `F`, such that `F=\GF{q}`, the `[n, k, d]` code
with length `n=\frac{q^{r}-1}{q-1}`, dimension `k=\frac{q^{r}-1}{q-1} - r` and
minimum distance `d=3` is called the Hamming Code of order `r`.

REFERENCES:

- [Rot2006]_
"""
# ****************************************************************************
#       Copyright (C) 2016 David Lucas <david.lucas@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from .linear_code import AbstractLinearCode
from sage.matrix.matrix_space import MatrixSpace
from sage.schemes.projective.projective_space import ProjectiveSpace
from sage.misc.cachefunc import cached_method
from sage.rings.integer import Integer


class HammingCode(AbstractLinearCode):
    r"""
    Representation of a Hamming code.

    INPUT:

    - ``base_field`` -- the base field over which ``self`` is defined

    - ``order`` -- the order of ``self``

    EXAMPLES::

        sage: C = codes.HammingCode(GF(7), 3)
        sage: C
        [57, 54] Hamming Code over GF(7)
    """
    _registered_encoders = {}
    _registered_decoders = {}

    def __init__(self, base_field, order):
        r"""
        TESTS:

        If ``base_field`` is not a finite field, an exception is raised::

            sage: codes.HammingCode(RR, 3)
            Traceback (most recent call last):
            ...
            ValueError: base_field has to be a finite field

        If ``order`` is not a Sage Integer or a Python int, an exception is raised::

            sage: codes.HammingCode(GF(3), 3.14)
            Traceback (most recent call last):
            ...
            ValueError: order has to be a Sage Integer or a Python int
        """
        if not base_field.is_finite():
            raise ValueError("base_field has to be a finite field")
        if not isinstance(order, (Integer, int)):
            raise ValueError("order has to be a Sage Integer or a Python int")

        q = base_field.order()
        length = Integer((q ** order - 1) / (q - 1))
        super().__init__(base_field, length, "Systematic", "Syndrome")
        self._dimension = length - order

    def __eq__(self, other):
        r"""
        Test equality of Hamming Code objects.

        EXAMPLES::

            sage: C1 = codes.HammingCode(GF(7), 3)
            sage: C2 = codes.HammingCode(GF(7), 3)
            sage: C1 == C2
            True
        """
        return isinstance(other, HammingCode)\
                and self.length() == other.length()\
                and self.dimension() == other.dimension()

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: C1 = codes.HammingCode(GF(7), 3)
            sage: C2 = codes.HammingCode(GF(7), 3)
            sage: hash(C1) == hash(C2)
            True
        """
        return hash((self.length(), self.dimension()))

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: C = codes.HammingCode(GF(7), 3)
            sage: C
            [57, 54] Hamming Code over GF(7)
        """
        return "[%s, %s] Hamming Code over GF(%s)"\
                % (self.length(), self.dimension(), self.base_field().cardinality())

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: C = codes.HammingCode(GF(7), 3)
            sage: latex(C)
            [57, 54] \textnormal{ Hamming Code over }\Bold{F}_{7}
        """
        return "[%s, %s] \\textnormal{ Hamming Code over }%s"\
                % (self.length(), self.dimension(), self.base_field()._latex_())

    @cached_method
    def parity_check_matrix(self):
        r"""
        Return a parity check matrix of ``self``.

        The construction of the parity check matrix in case ``self``
        is not a binary code is not really well documented.
        Regarding the choice of projective geometry, one might check:

        - the note over section 2.3 in [Rot2006]_, pages 47-48
        - the dedicated paragraph in [HP2003]_, page 30

        EXAMPLES::

            sage: C = codes.HammingCode(GF(3), 3)
            sage: C.parity_check_matrix()
            [1 0 1 1 0 1 0 1 1 1 0 1 1]
            [0 1 1 2 0 0 1 1 2 0 1 1 2]
            [0 0 0 0 1 1 1 1 1 2 2 2 2]
        """
        n = self.length()
        F = self.base_field()
        m = n - self.dimension()
        MS = MatrixSpace(F, n, m)
        X = ProjectiveSpace(m - 1, F)
        PFn = [list(p) for p in X.point_set(F).points()]

        H = MS(PFn).transpose()
        H = H[::-1, :]
        H.set_immutable()
        return H

    def minimum_distance(self):
        r"""
        Return the minimum distance of ``self``.

        It is always 3 as ``self`` is a Hamming Code.

        EXAMPLES::

            sage: C = codes.HammingCode(GF(7), 3)
            sage: C.minimum_distance()
            3
        """
        return 3



r"""
Brent Yorgey's fast algorithm for integer vector (multiset) partitions

ALGORITHM:

Brent Yorgey, Generating Multiset Partitions, The Monad Reader, Issue 8,
September 2007, p. 5.

https://wiki.haskell.org/The_Monad.Reader/Previous_issues

AUTHORS:

- D\. K\. Sunko (2020-02-19): initial version
- F\. Chapoton (2020-02-22): conversion to iterators and shorter doctests and
  doc tweaks
- T\. Scrimshaw (2020-03-06): Cython optimizations and doc tweaks
"""
################################################################################
#          Copyright (C) 2020 Denis Sunko <dks@phy.hr>                         #
#          Copyright (C) 2020 Frédéric Chapoton <chapoton@math.unistra.fr>     #
#          Copyright (C) 2020 Travis Scrimshaw <tscrims@gmail.com>             #
#                                                                              #
# This program is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by         #
# the Free Software Foundation, either version 2 of the License, or            #
# (at your option) any later version. The text of the license is found at      #
#                  https://www.gnu.org/licenses/                               #
################################################################################
#
# To understand the code below, consult the ALGORITHM.

cdef list vector_halve(list v):
    r"""
    Return the vector halfway (lexicographically) between ``v`` and zero.

    Internal part of :func:`fast_vector_partitions`.

    INPUT:

    - ``v`` -- list of nonnegative integers, understood as a vector

    OUTPUT:

    A list, understood as the integer vector halfway down the list of
    lexicographically ordered vectors between ``v`` and zero.

    EXAMPLES::

        sage: from sage.combinat.fast_vector_partitions import vector_halve  # not tested
        sage: vector_halve([1, 2, 3, 4, 5, 6, 7, 8, 9])  # not tested
        [0, 2, 3, 4, 5, 6, 7, 8, 9]
        sage: vector_halve([2, 4, 6, 8, 5, 6, 7, 8, 9])  # not tested
        [1, 2, 3, 4, 2, 6, 7, 8, 9]

    .. NOTE::

        For vectors, ``v = a + b`` implies ``v = b + a``, which means that a
        downward search for such splittings, starting with ``v = v + 0``, need
        only look as far as some "v / 2", given precise meaning here.

        A similar logic is to stop the search for divisors of ``N`` at
        ``sqrt(N)``, halving the exponents in the prime decomposition.
        However, here "v / 2" does not mean halving each coordinate.
    """
    cdef list result = list(v)  # make a copy
    cdef Py_ssize_t i, vv
    for i in range(len(v)):
        vv = <Py_ssize_t> v[i]
        result[i] = vv // 2
        if vv % 2:
            # the less significant part is just copied
            return result
    return result


def recursive_within_from_to(list m, list s, list e, bint useS, bint useE):
    r"""
    Iterate over a lexicographically ordered list of lists ``v`` satisfying
    ``e <= v <= s`` and ``v <|= m`` as vectors.

    Internal part of :func:`fast_vector_partitions`.

    INPUT:

    - ``m`` -- list of nonnegative integers, understood as a vector
    - ``s`` -- list of nonnegative integers, understood as a vector
    - ``e`` -- list of nonnegative integers, understood as a vector
    - ``useS`` -- boolean
    - ``useE`` -- boolean

    EXAMPLES::

        sage: from sage.combinat.fast_vector_partitions import recursive_within_from_to
        sage: list(recursive_within_from_to([1, 2, 3],[1, 2, 2],[1, 1, 1],True,True))
        [[1, 2, 2], [1, 2, 1], [1, 2, 0], [1, 1, 3], [1, 1, 2], [1, 1, 1]]

    .. NOTE::

        The flags ``useS`` and ``useE`` are used to implement the condition
        efficiently. Because testing it loops over the vector, re-testing
        at each step as the vector is parsed is inefficient: all but the last
        comparison have been done cumulatively already. This code tests
        only for the last one, using the flags to accumulate information
        from previous calls.

    .. WARNING::

        Expects to be called with ``s <|= m``.

        Expects to be called first with ``useS == useE == True``.
    """
    cdef Py_ssize_t start, end, x
    cdef bint useSS, useEE

    if useS:
        start = <Py_ssize_t> s[0]
    else:
        start = <Py_ssize_t> m[0]

    if useE:
        end = <Py_ssize_t> e[0]
    else:
        end = 0

    if len(m) == 1:
        # We use this style of Cython code for now in order to get this to convert
        #   to an optimized pure C for loop. See Cython github issue #532.
        # for x in range(start, end - 1, -1):
        for x from start >= x >= end by 1:
            yield [x]  # we know the answer for singletons
    else:
        # We use this style of Cython code for now in order to get this to convert
        #   to an optimized pure C for loop. See Cython github issue #532.
        # for x in range(start, end - 1, -1):
        for x from start >= x >= end by 1:
            useSS = useS and x == <Py_ssize_t> s[0]
            useEE = useE and x == <Py_ssize_t> e[0]
            for o in recursive_within_from_to(m[1:], s[1:], e[1:], useSS, useEE):
                yield [x] + o


def within_from_to(list m, list s, list e):
    r"""
    Iterate over a lexicographically ordered list of lists ``v`` satisfying
    ``e <= v <= s`` and ``v <|= m`` as vectors.

    Internal part of :func:`fast_vector_partitions`.

    INPUT:

    - ``m`` -- list of nonnegative integers, understood as a vector
    - ``s`` -- list of nonnegative integers, understood as a vector
    - ``e`` -- list of nonnegative integers, understood as a vector

    EXAMPLES::

        sage: from sage.combinat.fast_vector_partitions import within_from_to
        sage: list(within_from_to([1, 2, 3], [1, 2, 2], [1, 1, 1]))
        [[1, 2, 2], [1, 2, 1], [1, 2, 0], [1, 1, 3], [1, 1, 2], [1, 1, 1]]

    .. NOTE::

        The input ``s`` will be "clipped" internally if it does not satisfy
        the condition ``s <|= m``.

        To understand the input check, some line art is helpful. Assume
        that ``(a,b)`` are the two least significant coordinates of some
        vector. Say::

            e = (2,3), s = (7,6), m = (9,8).

        In the figure, these values are denoted by ``E``, ``S``, and ``M``,
        while the letter ``X`` stands for all other allowed values of
        ``v = (a,b)``::

            b ^
              |
            8 --------X---X---X---X---X-----------M
              |                                   |
            7 -       X   X   X   X   X           |
              |                                   |
            6 -       X   X   X   X   X   S       |
              |                                   |
            5 -       X   X   X   X   X   X       |
              |                                   |
            4 -       X   X   X   X   X   X       |
              |                                   |
            3 -       E   X   X   X   X   X       |
              |                                   |
            2 -           X   X   X   X   X       |
              |                                   |
            1 -           X   X   X   X   X       |
              |                                   |
            0 ----|---|---X---X---X---X---X---|---|--->
              0   1   2   3   4   5   6   7   8   9   a

        If ``S`` moves horizontally, the full-height columns fill the box in
        until ``S`` reaches ``M``, at which point it remains the limit in the
        b-direction as it moves out of the box, while M takes over as the
        limit in the a-direction, so the ``M``-column remains filled only up to
        ``S``, no matter how much ``S`` moves further to the right.

        If ``S`` moves vertically, its column will be filled to the top of the
        box, but it remains the relevant limit in the a-direction, while ``M``
        takes over in the b-direction as ``S`` goes out of the box upwards.

        Both behaviors are captured by using the smaller coordinate of ``S``
        and ``M``, whenever ``S`` is outside the box defined by M. The input
        will be "clipped" accordingly in that case.

    .. WARNING::

        The "clipping" behavior is transparent to the user, but may be
        puzzling when comparing outputs with the function
        :func:`recursive_within_from_to` which has no input protection.
    """
    cdef list ss = s
    # if s is not in the box defined by m, we must clip:
    cdef Py_ssize_t i, j
    for i in range(len(m)):  # should have the same length as s
        if s[i] > m[i]:
            ss = list(ss)  # make a copy
            ss[i] = m[i]
            for j in range(i+1, len(m)):
                if ss[j] > m[j]:
                    ss[j] = m[j]
            break
    if e > ss:
        return
    yield from recursive_within_from_to(m, ss, e, True, True)


cdef inline list vector_sub(list a, list b):
    """
    Return ``a - b`` considered as vectors.

    This assumes ``len(b) >= len(a)``.
    """
    cdef Py_ssize_t i
    cdef list ret = []
    for i in range(len(a)):
        ret.append((<Py_ssize_t> a[i]) - (<Py_ssize_t> b[i]))
    return ret


def recursive_vector_partitions(list v, list vL):
    r"""
    Iterate over a lexicographically ordered list of lists, each list
    representing a vector partition of ``v``, such that no part of any
    partition is lexicographically smaller than ``vL``.

    Internal part of :func:`fast_vector_partitions`.

    INPUT:

    - ``v`` -- list of nonnegative integers, understood as a vector
    - ``vL`` -- list of nonnegative integers, understood as a vector

    EXAMPLES::

        sage: from sage.combinat.fast_vector_partitions import recursive_vector_partitions
        sage: list(recursive_vector_partitions([2, 2, 2],[1, 1, 1]))
        [[[2, 2, 2]], [[1, 1, 1], [1, 1, 1]]]
        sage: list(recursive_vector_partitions([2, 2, 2],[1, 1, 0]))
        [[[2, 2, 2]], [[1, 1, 1], [1, 1, 1]], [[1, 1, 0], [1, 1, 2]]]
        sage: list(recursive_vector_partitions([2, 2, 2],[1, 0, 1]))
        [[[2, 2, 2]],
         [[1, 1, 1], [1, 1, 1]],
         [[1, 1, 0], [1, 1, 2]],
         [[1, 0, 2], [1, 2, 0]],
         [[1, 0, 1], [1, 2, 1]]]
    """
    cdef list v_minus_vv, pp, vv
    yield [v]
    for vv in within_from_to(v, vector_halve(v), vL):
        v_minus_vv = vector_sub(v, vv)
        for pp in recursive_vector_partitions(v_minus_vv, vv):
            yield [vv] + pp


def fast_vector_partitions(v, min_vals=None):
    r"""
    Brent Yorgey's fast algorithm for integer vector (multiset) partitions.

    INPUT:

    - ``v`` -- list of nonnegative integers, understood as the vector
      to be partitioned

    - ``min_vals`` -- (optional) list of nonnegative integers, of same
      length as ``v``

    OUTPUT: list of lists, each representing a vector partition of ``v``

    If ``min_vals`` is given, only partitions with parts ``p >= min_vals`` in
    the lexicographic ordering will appear.

    If ``min_vals`` is given and ``len(min_vals) != len(v)``, an error
    is raised.

    EXAMPLES:

    The older the computer, the more impressive the comparison::

        sage: from sage.combinat.fast_vector_partitions import fast_vector_partitions
        sage: fastvparts = list(fast_vector_partitions([3, 3, 3]))
        sage: vparts = list(VectorPartitions([3, 3, 3]))
        sage: vparts == fastvparts[::-1]
        True
        sage: len(fastvparts)
        686
        sage: list(fast_vector_partitions([1, 2, 3], min_vals=[0, 1, 1]))
        [[[1, 2, 3]],
         [[0, 2, 3], [1, 0, 0]],
         [[0, 2, 2], [1, 0, 1]],
         [[0, 2, 1], [1, 0, 2]],
         [[0, 2, 0], [1, 0, 3]],
         [[0, 1, 3], [1, 1, 0]],
         [[0, 1, 2], [1, 1, 1]],
         [[0, 1, 1], [1, 1, 2]],
         [[0, 1, 1], [0, 1, 2], [1, 0, 0]],
         [[0, 1, 1], [0, 1, 1], [1, 0, 1]]]
        sage: L1 = list(fast_vector_partitions([5, 7, 6], min_vals=[1, 3, 2]))
        sage: L1 == list(VectorPartitions([5, 7, 6], min=[1, 3, 2]))[::-1]
        True

    .. NOTE::

        The partitions are returned as an iterator.

        In this documentation, ``a <|= b`` means ``a[i] <= b[i]`` for all ``i``
        (notation following B. Yorgey's paper). It is the monomial partial
        ordering in Dickson's lemma: ``a <|= b`` iff ``x^a`` divides ``x^b`` as
        monomials.

    .. WARNING::

        The ordering of the partitions is reversed with respect to the output of
        Sage class :class:`~sage.combinat.vector_partition.VectorPartitions`.
    """
    if min_vals is None:
        min_vals = (len(v) - 1) * [0] + [1]  # lexicographically smallest vector > 0
    if len(v) != len(min_vals):
        raise ValueError("the length of v and min_vals must be equal")
    return recursive_vector_partitions(list(v), list(min_vals))



# sage.doctest: needs sage.combinat sage.geometry.polyhedron
r"""
Associahedron

.. TODO::

    - fix adjacency matrix
    - edit graph method to get proper vertex labellings
    - UniqueRepresentation?

AUTHORS:

- Christian Stump
"""
# ***************************************************************************
#       Copyright (C) 2011-2012 Christian Stump <christian.stump@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ***************************************************************************
from sage.geometry.polyhedron.backend_ppl import Polyhedron_QQ_ppl
from sage.geometry.polyhedron.backend_normaliz import Polyhedron_QQ_normaliz
from sage.geometry.polyhedron.backend_cdd import Polyhedron_QQ_cdd
from sage.geometry.polyhedron.backend_field import Polyhedron_field
from sage.geometry.polyhedron.backend_polymake import Polyhedron_polymake
from sage.geometry.polyhedron.parent import Polyhedra, Polyhedra_base, Polyhedra_QQ_ppl, Polyhedra_QQ_normaliz, Polyhedra_QQ_cdd, Polyhedra_polymake, Polyhedra_field
from sage.combinat.root_system.cartan_type import CartanType
from sage.modules.free_module_element import vector
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

ancestors_of_associahedron = set([Polyhedron_QQ_ppl, Polyhedron_QQ_normaliz, Polyhedron_QQ_cdd, Polyhedron_field, Polyhedron_polymake])


def Associahedron(cartan_type, backend='ppl'):
    r"""
    Construct an associahedron.

    The generalized associahedron is a polytopal complex with vertices in
    one-to-one correspondence with clusters in the cluster complex, and with
    edges between two vertices if and only if the associated two clusters
    intersect in codimension 1.

    The associahedron of type `A_n` is one way to realize the classical
    associahedron as defined in the :wikipedia:`Associahedron`.

    A polytopal realization of the associahedron can be found in [CFZ2002]_. The
    implementation is based on [CFZ2002]_, Theorem 1.5, Remark 1.6, and Corollary
    1.9.

    INPUT:

    - ``cartan_type`` -- a cartan type according to
      :class:`sage.combinat.root_system.cartan_type.CartanTypeFactory`

    - ``backend`` -- string (``'ppl'``); the backend to use;
      see :meth:`sage.geometry.polyhedron.constructor.Polyhedron`

    EXAMPLES::

        sage: Asso = polytopes.associahedron(['A',2]); Asso
        Generalized associahedron of type ['A', 2] with 5 vertices

        sage: sorted(Asso.Hrepresentation(), key=repr)
        [An inequality (-1, 0) x + 1 >= 0,
         An inequality (0, -1) x + 1 >= 0,
         An inequality (0, 1) x + 1 >= 0,
         An inequality (1, 0) x + 1 >= 0,
         An inequality (1, 1) x + 1 >= 0]

        sage: Asso.Vrepresentation()
        (A vertex at (1, -1), A vertex at (1, 1), A vertex at (-1, 1),
         A vertex at (-1, 0), A vertex at (0, -1))

        sage: polytopes.associahedron(['B',2])
        Generalized associahedron of type ['B', 2] with 6 vertices

    The two pictures of [CFZ2002]_ can be recovered with::

        sage: Asso = polytopes.associahedron(['A',3]); Asso
        Generalized associahedron of type ['A', 3] with 14 vertices
        sage: Asso.plot()  # long time
        Graphics3d Object

        sage: Asso = polytopes.associahedron(['B',3]); Asso
        Generalized associahedron of type ['B', 3] with 20 vertices
        sage: Asso.plot()  # long time
        Graphics3d Object

    TESTS::

        sage: sorted(polytopes.associahedron(['A',3]).vertices())
        [A vertex at (-3/2, 0, -1/2), A vertex at (-3/2, 0, 3/2),
         A vertex at (-3/2, 1, -3/2), A vertex at (-3/2, 2, -3/2),
         A vertex at (-3/2, 2, 3/2), A vertex at (-1/2, -1, -1/2),
         A vertex at (-1/2, 0, -3/2), A vertex at (1/2, -2, 1/2),
         A vertex at (1/2, -2, 3/2), A vertex at (3/2, -2, 1/2),
         A vertex at (3/2, -2, 3/2), A vertex at (3/2, 0, -3/2),
         A vertex at (3/2, 2, -3/2), A vertex at (3/2, 2, 3/2)]

        sage: sorted(polytopes.associahedron(['B',3]).vertices())
        [A vertex at (-3, 0, 0), A vertex at (-3, 0, 3),
         A vertex at (-3, 2, -2), A vertex at (-3, 4, -3),
         A vertex at (-3, 5, -3), A vertex at (-3, 5, 3),
         A vertex at (-2, 1, -2), A vertex at (-2, 3, -3),
         A vertex at (-1, -2, 0), A vertex at (-1, -1, -1),
         A vertex at (1, -4, 1), A vertex at (1, -3, 0),
         A vertex at (2, -5, 2), A vertex at (2, -5, 3),
         A vertex at (3, -5, 2), A vertex at (3, -5, 3),
         A vertex at (3, -3, 0), A vertex at (3, 3, -3),
         A vertex at (3, 5, -3), A vertex at (3, 5, 3)]

        sage: polytopes.associahedron(['A',4]).f_vector()
        (1, 42, 84, 56, 14, 1)
        sage: polytopes.associahedron(['B',4]).f_vector()
        (1, 70, 140, 90, 20, 1)

        sage: p1 = polytopes.associahedron(['A',4], backend='normaliz')   # optional - pynormaliz
        sage: TestSuite(p1).run(skip='_test_pickling')                    # optional - pynormaliz
        sage: p2 = polytopes.associahedron(['A',4], backend='cdd')
        sage: TestSuite(p2).run()
        sage: p3 = polytopes.associahedron(['A',4], backend='field')
        sage: TestSuite(p3).run()
    """
    cartan_type = CartanType(cartan_type)
    parent = Associahedra(QQ, cartan_type.rank(), backend)
    return parent(cartan_type)


class Associahedron_class_base:
    r"""
    The base class of the Python class of an associahedron

    You should use the :func:`Associahedron` convenience function to
    construct associahedra from the Cartan type.

    TESTS::

        sage: Asso = polytopes.associahedron(['A',2]); Asso
        Generalized associahedron of type ['A', 2] with 5 vertices
        sage: TestSuite(Asso).run()
    """
    def __new__(typ, parent=None, Vrep=None, Hrep=None, cartan_type=None, **kwds):
        r"""
        Return instance of :class:`Assciahedron_class_base`, if ``cartan_type`` is provided
        or object is being unpickled.

        In other cases, this call is a result of a polyhedral construction with an associahedron.
        Thus we return the corresponding instance of
        :class:`sage.geometry.polyhedron.base.Polyhedron_base` (not an associahedron).

        TESTS:

        Check that faces of associahedra work::

            sage: A = polytopes.associahedron(['A',3], backend='ppl'); A
            Generalized associahedron of type ['A', 3] with 14 vertices
            sage: face = A.faces(2)[3]
            sage: P = face.as_polyhedron(); P
            A 2-dimensional polyhedron in QQ^3 defined as the convex hull of 4 vertices
            sage: P.backend()
            'ppl'
            sage: A = polytopes.associahedron(['A',3], backend='field'); A
            Generalized associahedron of type ['A', 3] with 14 vertices
            sage: A.faces(2)[3].as_polyhedron().backend()
            'field'

        Check other polytopal constructions::

            sage: A = polytopes.associahedron(['A',4], backend='ppl')
            sage: A + A
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 42 vertices
            sage: A - A
            A 0-dimensional polyhedron in QQ^4 defined as the convex hull of 1 vertex
            sage: A.intersection(A)
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 42 vertices
            sage: A.translation(A.center())
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 42 vertices
            sage: A.dilation(2)
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 42 vertices
            sage: A.dilation(2.0)
            A 4-dimensional polyhedron in RDF^4 defined as the convex hull of 42 vertices
            sage: A.convex_hull(A)
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 42 vertices
            sage: A.polar()
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 14 vertices
        """
        if cartan_type or (parent is None and Vrep is None and Hrep is None):
            # Called from element constructor in ``Associahedron_base``.
            # Alternatively called from ``loads`` in ``loads(dumps(...))``.
            return super().__new__(typ, parent, Vrep, Hrep, **kwds)
        # Not called from element constructor in ``Associahedron_base``.
        # Return a polyhedron with proper backend (not an associahedron).
        # Thus e.g. a face of an Associahedron can be initialized as a polyhedron.
        mro = typ.mro()
        for typ1 in mro:
            if typ1 in ancestors_of_associahedron:
                return typ1(parent, Vrep, Hrep, **kwds)
        raise ValueError("could not determine a parent class")

    def __init__(self, parent, Vrep, Hrep, cartan_type=None, **kwds):
        r"""
        Initialize an associahedron.

        If ``'cartan_type'`` is ``None``, :meth:`Associahedron_class_base.__new__`
        returns a (general) polyhedron instead.

        TESTS::

            sage: A = polytopes.associahedron(['A',3], backend='ppl'); A
            Generalized associahedron of type ['A', 3] with 14 vertices
        """
        if cartan_type:
            self._cartan_type = cartan_type
            super().__init__(parent, Vrep, Hrep, **kwds)
        else:
            raise ValueError("associahedron must be initialized with cartan type")

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: polytopes.associahedron(['A',3])._repr_()
            "Generalized associahedron of type ['A', 3] with 14 vertices"
        """
        msg = 'Generalized associahedron of type {} with {} vertices'
        return msg.format(self._cartan_type, self.n_vertices())

    def cartan_type(self):
        r"""
        Return the Cartan type of ``self``.

        EXAMPLES::

            sage: polytopes.associahedron(['A',3]).cartan_type()
            ['A', 3]
        """
        return self._cartan_type

    def vertices_in_root_space(self):
        r"""
        Return the vertices of ``self`` as elements in the root space.

        EXAMPLES::

            sage: Asso = polytopes.associahedron(['A',2])
            sage: Asso.vertices()
            (A vertex at (1, -1), A vertex at (1, 1),
             A vertex at (-1, 1), A vertex at (-1, 0),
             A vertex at (0, -1))

            sage: Asso.vertices_in_root_space()
            (alpha[1] - alpha[2], alpha[1] + alpha[2], -alpha[1] + alpha[2],
            -alpha[1], -alpha[2])
        """
        root_space = self._cartan_type.root_system().root_space()
        return tuple(root_space.from_vector(vector(V))
                     for V in self.vertex_generator())


class Associahedron_class_ppl(Associahedron_class_base, Polyhedron_QQ_ppl):
    pass


class Associahedron_class_normaliz(Associahedron_class_base, Polyhedron_QQ_normaliz):
    pass


class Associahedron_class_cdd(Associahedron_class_base, Polyhedron_QQ_cdd):
    pass


class Associahedron_class_polymake(Associahedron_class_base, Polyhedron_polymake):
    pass


class Associahedron_class_field(Associahedron_class_base, Polyhedron_field):
    pass


def Associahedra(base_ring, ambient_dim, backend='ppl'):
    r"""
    Construct a parent class of Associahedra according to ``backend``.

    TESTS::

        sage: from sage.combinat.root_system.associahedron import Associahedra
        sage: Associahedra(QQ, 4, 'ppl').parent()
        <class 'sage.combinat.root_system.associahedron.Associahedra_ppl_with_category'>
        sage: Associahedra(QQ, 4, 'normaliz').parent() # optional - pynormaliz
        <class 'sage.combinat.root_system.associahedron.Associahedra_normaliz_with_category'>
        sage: Associahedra(QQ, 4, 'polymake').parent() # optional - jupymake
        <class 'sage.combinat.root_system.associahedron.Associahedra_polymake_with_category'>
        sage: Associahedra(QQ, 4, 'field').parent()
        <class 'sage.combinat.root_system.associahedron.Associahedra_field_with_category'>
        sage: Associahedra(QQ, 4, 'cdd').parent()
        <class 'sage.combinat.root_system.associahedron.Associahedra_cdd_with_category'>

    .. SEEALSO::

        :class:`Associahedra_base`.
    """
    if base_ring is not QQ:
        raise NotImplementedError("base ring must be QQ")
    if backend == 'ppl':
        return Associahedra_ppl(base_ring, ambient_dim, backend)
    if backend == 'normaliz':
        return Associahedra_normaliz(base_ring, ambient_dim, backend)
    if backend == 'cdd':
        return Associahedra_cdd(QQ, ambient_dim, backend)
    if backend == 'polymake':
        return Associahedra_polymake(base_ring.fraction_field(), ambient_dim, backend)
    if backend == 'field':
        return Associahedra_field(base_ring, ambient_dim, backend)
    raise ValueError("unknown backend")


class Associahedra_base:
    """
    Base class of parent of Associahedra of specified dimension

    EXAMPLES::

        sage: from sage.combinat.root_system.associahedron import Associahedra
        sage: parent = Associahedra(QQ,2,'ppl');  parent
        Polyhedra in QQ^2
        sage: type(parent)
        <class 'sage.combinat.root_system.associahedron.Associahedra_ppl_with_category'>
        sage: parent(['A',2])
        Generalized associahedron of type ['A', 2] with 5 vertices

    Importantly, the parent knows the dimension of the ambient
    space. If you try to construct an associahedron of a different
    dimension, a :exc:`ValueError` is raised::

        sage: parent(['A',3])
        Traceback (most recent call last):
        ...
        ValueError: V-representation data requires a list of length ambient_dim
    """

    def _element_constructor_(self, cartan_type, **kwds):
        """
        The element constructor.

        This method is called internally when we try to convert
        something into an element. In this case, the only thing that
        can be converted into an associahedron is the Cartan type.

        EXAMPLES::

            sage: from sage.combinat.root_system.associahedron import Associahedra
            sage: parent = Associahedra(QQ,2,'ppl')
            sage: parent(['A',2])
            Generalized associahedron of type ['A', 2] with 5 vertices
            sage: parent._element_constructor_(['A',2])
            Generalized associahedron of type ['A', 2] with 5 vertices

        TESTS::

            sage: parent(['A', 2, 1])
            Traceback (most recent call last):
            ...
            ValueError: the Cartan type must be finite
        """
        cartan_type = CartanType(cartan_type)
        if not cartan_type.is_finite():
            raise ValueError("the Cartan type must be finite")
        root_space = cartan_type.root_system().root_space()
        # TODO: generalize this as a method of root lattice realization
        rhocheck = sum(beta.associated_coroot()
                       for beta in root_space.positive_roots()) / 2
        I = root_space.index_set()
        inequalities = []
        for orbit in root_space.almost_positive_roots_decomposition():
            c = rhocheck.coefficient(orbit[0].leading_support())
            for beta in orbit:
                inequalities.append([c] + [beta.coefficient(i) for i in I])
        associahedron = super()._element_constructor_(None, [inequalities, []], cartan_type=cartan_type)
        return associahedron

    def _coerce_map_from_(self, X):
        r"""
        Return whether there is a coercion from ``X``.

        INPUT:

        - ``X`` -- anything

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.geometry.polyhedron.parent import Polyhedra
            sage: from sage.combinat.root_system.associahedron import Associahedra
            sage: Associahedra(QQ,3).has_coerce_map_from( Polyhedra(QQ,3) )   # indirect doctest
            False
            sage: Polyhedra(QQ,3).has_coerce_map_from( Associahedra(QQ,3) )
            True

        TESTS::

            sage: A = polytopes.associahedron(['A',4], backend='ppl'); type(A.parent())
            <class 'sage.combinat.root_system.associahedron.Associahedra_ppl_with_category'>
            sage: B = polytopes.simplex().change_ring(QQ); type(B.parent())
            <class 'sage.geometry.polyhedron.parent.Polyhedra_QQ_ppl_with_category'>
            sage: A + B
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 70 vertices
            sage: A - B
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 24 vertices
            sage: A.intersection(B)
            A 3-dimensional polyhedron in QQ^4 defined as the convex hull of 4 vertices
            sage: A.convex_hull(B)
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 42 vertices
        """
        if not isinstance(X, Associahedra_base):
            return False
        return super()._coerce_map_from_(X)

    def _pushout_(self, other):
        r"""
        The pushout of Polyhedra over ZZ and Associahedra over QQ is Polyhedra over QQ.

        TESTS::

            sage: A = polytopes.associahedron(['A',4], backend='ppl'); type(A.parent())
            <class 'sage.combinat.root_system.associahedron.Associahedra_ppl_with_category'>
            sage: B = polytopes.simplex(); type(B.parent())
            <class 'sage.geometry.polyhedron.parent.Polyhedra_ZZ_ppl_with_category'>
            sage: A + B
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 70 vertices
            sage: A - B
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 24 vertices
            sage: A.intersection(B)
            A 3-dimensional polyhedron in QQ^4 defined as the convex hull of 4 vertices
            sage: A.convex_hull(B)
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 42 vertices
        """
        if isinstance(other, Polyhedra_base) and other.base_ring() == ZZ:
            return Polyhedra(QQ, self.ambient_dim(), self.backend())

        # Call the overwritten pushout in case it exists.
        if hasattr(super(), '_pushout_'):
            return super()._pushout_(other)


class Associahedra_ppl(Associahedra_base, Polyhedra_QQ_ppl):
    Element = Associahedron_class_ppl


class Associahedra_normaliz(Associahedra_base, Polyhedra_QQ_normaliz):
    Element = Associahedron_class_normaliz


class Associahedra_cdd(Associahedra_base, Polyhedra_QQ_cdd):
    Element = Associahedron_class_cdd


class Associahedra_polymake(Associahedra_base, Polyhedra_polymake):
    Element = Associahedron_class_polymake


class Associahedra_field(Associahedra_base, Polyhedra_field):
    Element = Associahedron_class_field



from typing import Any

from sage.libs.mpc.types import mpc_t, mpc_rnd_t
from sage.structure.element import FieldElement
from sage.rings.ring import Field

class MPComplexNumber(FieldElement):
    value: mpc_t
    init: bool

    def _new(self) -> 'MPComplexNumber':
        ...

    def _add_(self, other: Any) -> Any:
        ...

    def _mul_(self, other: Any) -> Any:
        ...

class MPComplexField_class(Field):
    _prec: int
    __rnd: mpc_rnd_t
    __rnd_str: Any
    __real_field: Any
    __imag_field: Any

    def _new(self) -> MPComplexNumber:
        ...

    def _an_element_(self) -> MPComplexNumber:
        ...



from __future__ import annotations
from collections.abc import Callable
from typing import Any

from sage.symbolic.function import SymbolicFunction

def function_factory(
    name: str,
    nargs: int = 0,
    latex_name: str | None = None,
    conversions: dict[str, Any] | None = None,
    evalf_params_first: bool = True,
    eval_func: Callable | None = None,
    evalf_func: Callable | None = None,
    conjugate_func: Callable | None = None,
    real_part_func: Callable | None = None,
    imag_part_func: Callable | None = None,
    derivative_func: Callable | None = None,
    tderivative_func: Callable | None = None,
    power_func: Callable | None = None,
    series_func: Callable | None = None,
    print_func: Callable | None = None,
    print_latex_func: Callable | None = None,
) -> SymbolicFunction:
    ...

def unpickle_function(
    name: str,
    nargs: int,
    latex_name: str | None,
    conversions: dict[str, Any] | None,
    evalf_params_first: bool,
    pickled_funcs: list[Any],
) -> SymbolicFunction:
    ...

def function(
    s: str,
    **kwds: Any
) -> SymbolicFunction | list[SymbolicFunction]:
    ...



# sage.doctest: needs sage.libs.flint
"""
`L`-series of modular abelian varieties

AUTHOR:

- William Stein (2007-03)

TESTS::

    sage: L = J0(37)[0].padic_lseries(5)
    sage: loads(dumps(L)) == L
    True
    sage: L = J0(37)[0].lseries()
    sage: loads(dumps(L)) == L
    True
"""

# #########################################################################
#       Copyright (C) 2007 William Stein <wstein@gmail.com>               #
#  Distributed under the terms of the GNU General Public License (GPL)    #
#                  https://www.gnu.org/licenses/                           #
# #########################################################################

from sage.structure.sage_object import SageObject
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.rings.integer import Integer
from sage.rings.infinity import infinity
from sage.rings.cc import CC
from sage.modules.free_module import span
from sage.misc.misc_c import prod


class Lseries(SageObject):
    """
    Base class for `L`-series attached to modular abelian varieties.

    This is a common base class for complex and `p`-adic `L`-series
    of modular abelian varieties.
    """
    def __init__(self, abvar):
        """
        Called when creating an `L`-series.

        INPUT:

        - ``abvar`` -- a modular abelian variety

        EXAMPLES::

            sage: J0(11).lseries()
            Complex L-series attached to Abelian variety J0(11) of dimension 1
            sage: J0(11).padic_lseries(7)
            7-adic L-series attached to Abelian variety J0(11) of dimension 1
        """
        self.__abvar = abvar

    def abelian_variety(self):
        """
        Return the abelian variety that this `L`-series is attached to.

        OUTPUT: a modular abelian variety

        EXAMPLES::

            sage: J0(11).padic_lseries(7).abelian_variety()
            Abelian variety J0(11) of dimension 1
        """
        return self.__abvar


class Lseries_complex(Lseries):
    """
    A complex `L`-series attached to a modular abelian variety.

    EXAMPLES::

        sage: A = J0(37)
        sage: A.lseries()
        Complex L-series attached to Abelian variety J0(37) of dimension 2
    """
    def __call__(self, s, prec=53):
        """
        Evaluate this complex `L`-series at `s`.

        INPUT:

        - ``s`` -- complex number

        - ``prec`` -- integer (default: 53); the number of bits of precision
          used in computing the lseries of the newforms

        OUTPUT: a complex number L(A, s)

        EXAMPLES::

            sage: L = J0(23).lseries()
            sage: L(1)                                                                  # needs sage.symbolic
            0.248431866590600
            sage: L(1, prec=100)                                                        # needs sage.symbolic
            0.24843186659059968120725033931

            sage: L = J0(389)[0].lseries()
            sage: L(1)                          # abstol 1e-10          # long time (2s), needs sage.symbolic
            -1.33139759782370e-19
            sage: L(1, prec=100)                # abstol 1e-20          # long time (2s), needs sage.symbolic
            6.0129758648142797032650287762e-39
            sage: L.rational_part()
            0

            sage: L = J1(23)[0].lseries()
            sage: L(1)                                                                  # needs sage.symbolic
            0.248431866590600

            sage: J = J0(11) * J1(11)
            sage: J.lseries()(1)                                                        # needs sage.symbolic
            0.0644356903227915

            sage: L = JH(17,[2]).lseries()
            sage: L(1)                                                                  # needs sage.symbolic
            0.386769938387780
        """
        abelian_variety = self.abelian_variety()
        # Check for easy dimension zero case
        if abelian_variety.dimension() == 0:
            return CC(1)
        try:
            factors = self.__factors[prec]
            return prod(L(s) for L in factors)
        except AttributeError:
            self.__factors = {}
        except KeyError:
            pass
        abelian_variety = self.abelian_variety()
        newforms = abelian_variety.newform_decomposition('a')

        factors = [newform.lseries(embedding=i, prec=prec)
                for newform in newforms
                for i in range(newform.base_ring().degree())]
        self.__factors[prec] = factors

        return prod(L(s) for L in factors)

    def __eq__(self, other):
        """
        Compare this complex `L`-series to another one.

        INPUT:

        - ``other`` -- object

        OUTPUT: boolean

        EXAMPLES::

            sage: L = J0(37)[0].lseries()
            sage: M = J0(37)[1].lseries()
            sage: L == M
            False
            sage: L == L
            True
        """
        if not isinstance(other, Lseries_complex):
            return False
        return self.abelian_variety() == other.abelian_variety()

    def __ne__(self, other):
        """
        Check whether ``self`` is not equal to ``other``.

        INPUT:

        - ``other`` -- object

        OUTPUT: boolean

        EXAMPLES::

            sage: L = J0(37)[0].lseries()
            sage: M = J0(37)[1].lseries()
            sage: L != M
            True
            sage: L != L
            False
        """
        return not (self == other)

    def _repr_(self):
        """
        String representation of `L`-series.

        OUTPUT: string

        EXAMPLES::

            sage: L = J0(37).lseries()
            sage: L._repr_()
            'Complex L-series attached to Abelian variety J0(37) of dimension 2'
        """
        return "Complex L-series attached to %s" % self.abelian_variety()

    def vanishes_at_1(self):
        """
        Return ``True`` if `L(1)=0` and return ``False`` otherwise.

        OUTPUT: boolean

        EXAMPLES:

        Numerically, the `L`-series for `J_0(389)` appears to vanish
        at 1.  This is confirmed by this algebraic computation::

            sage: L = J0(389)[0].lseries(); L
            Complex L-series attached to Simple abelian subvariety 389a(1,389) of dimension 1 of J0(389)
            sage: L(1) # long time (2s) abstol 1e-10
            -1.33139759782370e-19
            sage: L.vanishes_at_1()
            True

        Numerically, one might guess that the `L`-series for `J_1(23)`
        and `J_1(31)` vanish at 1.  This algebraic computation shows
        otherwise::

            sage: L = J1(23).lseries(); L
            Complex L-series attached to Abelian variety J1(23) of dimension 12
            sage: L(1)  # long time (about 3 s)
            0.0001295198...
            sage: L.vanishes_at_1()
            False
            sage: abs(L(1, prec=100)- 0.00012951986142702571478817757148) < 1e-32  # long time (about 3 s)
            True

            sage: L = J1(31).lseries(); L
            Complex L-series attached to Abelian variety J1(31) of dimension 26
            sage: abs(L(1) - 3.45014267547611e-7) < 1e-15  # long time (about 8 s)
            True
            sage: L.vanishes_at_1()  # long time (about 6 s)
            False
        """
        abelian_variety = self.abelian_variety()
        # Check for easy dimension zero case
        if abelian_variety.dimension() == 0:
            return False
        if not abelian_variety.is_simple():
            from .constructor import AbelianVariety
            decomp = (AbelianVariety(f) for f in
                      abelian_variety.newform_decomposition('a'))
            return any(S.lseries().vanishes_at_1() for S in decomp)
        modular_symbols = abelian_variety.modular_symbols()
        Phi = modular_symbols.rational_period_mapping()
        ambient_module = modular_symbols.ambient_module()

        e = ambient_module([0, infinity])
        return Phi(e).is_zero()

    def rational_part(self):
        """
        Return the rational part of this `L`-function at the central critical
        value 1.

        OUTPUT: a rational number

        EXAMPLES::

            sage: A, B = J0(43).decomposition()
            sage: A.lseries().rational_part()
            0
            sage: B.lseries().rational_part()
            2/7
        """
        abelian_variety = self.abelian_variety()
        modular_symbols = abelian_variety.modular_symbols()
        Phi = modular_symbols.rational_period_mapping()
        ambient_module = modular_symbols.ambient_module()

        if self.vanishes_at_1():
            return QQ(0)
        s = ambient_module.sturm_bound()
        I = ambient_module.hecke_images(0, range(1, s+1))
        PhiTe = span([Phi(ambient_module(I[n]))
            for n in range(I.nrows())], ZZ)

        ambient_plus = ambient_module.sign_submodule(1)
        ambient_plus_cusp = ambient_plus.cuspidal_submodule()
        PhiH1plus = span([Phi(x) for
            x in ambient_plus_cusp.integral_basis()], ZZ)

        return PhiTe.index_in(PhiH1plus)

    lratio = rational_part


class Lseries_padic(Lseries):
    """
    A `p`-adic `L`-series attached to a modular abelian variety.
    """
    def __init__(self, abvar, p):
        """
        Create a `p`-adic `L`-series.

        EXAMPLES::

            sage: J0(37)[0].padic_lseries(389)
            389-adic L-series attached to Simple abelian subvariety 37a(1,37) of dimension 1 of J0(37)
        """
        Lseries.__init__(self, abvar)
        p = Integer(p)
        if not p.is_prime():
            raise ValueError("p (=%s) must be prime" % p)
        self.__p = p

    def __eq__(self, other):
        """
        Compare this `p`-adic `L`-series to another one.

        First the abelian varieties are compared; if they are the same,
        then the primes are compared.

        INPUT:

        - ``other`` -- object

        OUTPUT: boolean

        EXAMPLES::

            sage: L = J0(37)[0].padic_lseries(5)
            sage: M = J0(37)[1].padic_lseries(5)
            sage: K = J0(37)[0].padic_lseries(3)
            sage: L == K
            False
            sage: L == M
            False
            sage: L == L
            True
        """
        if not isinstance(other, Lseries_padic):
            return False
        return (self.abelian_variety() == other.abelian_variety() and
                self.__p == other.__p)

    def __ne__(self, other):
        """
        Check whether ``self`` is not equal to ``other``.

        INPUT:

        - ``other`` -- object

        OUTPUT: boolean

        EXAMPLES::

            sage: L = J0(37)[0].padic_lseries(5)
            sage: M = J0(37)[1].padic_lseries(5)
            sage: K = J0(37)[0].padic_lseries(3)
            sage: L != K
            True
            sage: L != M
            True
            sage: L != L
            False
        """
        return not (self == other)

    def prime(self):
        """
        Return the prime `p` of this `p`-adic `L`-series.

        EXAMPLES::

            sage: J0(11).padic_lseries(7).prime()
            7
        """
        return self.__p

    def power_series(self, n=2, prec=5):
        """
        Return the `n`-th approximation to this `p`-adic `L`-series as
        a power series in `T`.

        Each coefficient is a `p`-adic number
        whose precision is provably correct.

        NOTE: This is not yet implemented.

        EXAMPLES::

            sage: L = J0(37)[0].padic_lseries(5)
            sage: L.power_series()
            Traceback (most recent call last):
            ...
            NotImplementedError
            sage: L.power_series(3,7)
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        raise NotImplementedError

    def _repr_(self):
        """
        String representation of this `p`-adic `L`-series.

        EXAMPLES::

            sage: L = J0(37)[0].padic_lseries(5)
            sage: L._repr_()
            '5-adic L-series attached to Simple abelian subvariety 37a(1,37) of dimension 1 of J0(37)'
        """
        return "%s-adic L-series attached to %s" % (self.__p,
                                                    self.abelian_variety())



from sage.libs.mpfr.types cimport mpfr_prec_t
from sage.libs.mpfi.types cimport mpfi_t

from sage.rings.ring cimport Field
cimport sage.rings.abc
from sage.structure.element cimport RingElement

from sage.rings.rational cimport Rational
from sage.rings.real_mpfr cimport RealField_class

cdef class RealIntervalFieldElement(RingElement)  # forward decl

cdef class RealIntervalField_class(sage.rings.abc.RealIntervalField):
    cdef mpfr_prec_t _prec
    cdef bint sci_not
    # Cache RealField instances for the lower, upper, and middle bounds.
    # These have the same precision as the interval field;
    # __lower_field rounds down, __upper_field rounds up.
    # These fields with their roundings are not used for computation
    # in this module, but they do affect the printing and the return
    # values of lower() and upper().  Consider a 3-bit
    # interval containing exactly the floating-point number 1.25.
    # In round-to-nearest or round-down, this prints as 1.2; in round-up,
    # this prints as 1.3.  The straightforward options, then, are to
    # print this interval as [1.2 ... 1.2] (which does not even contain
    # the true value, 1.25), or to print it as [1.2 ... 1.3] (which
    # gives the impression that the upper and lower bounds are not
    # equal, even though they really are).  Neither of these is very
    # satisfying, but I have chosen the latter for now.
    cdef RealField_class __lower_field
    cdef RealField_class __middle_field
    cdef RealField_class __upper_field
    cdef object _multiplicative_order

    cdef inline RealIntervalFieldElement _new(self):
        """Return a new real interval with parent ``self``."""
        t = <type>self.element_class
        return <RealIntervalFieldElement>(t.__new__(t, self))


cdef class RealIntervalFieldElement(RingElement):
    cdef mpfi_t value

    cdef inline RealIntervalFieldElement _new(self):
        """Return a new real interval with same parent as ``self``."""
        return (<RealIntervalField_class>self._parent)._new()
    cdef RealIntervalFieldElement abs(RealIntervalFieldElement self)
    cdef Rational _simplest_rational_helper(self)
    cpdef _str_question_style(self, int base, int error_digits, e, bint prefer_sci)



# sage.doctest: needs sage.symbolic
"""
Regular polygons in the upper half model for hyperbolic plane

AUTHORS:

- Javier Honrubia (2016-01)
"""

#******************************************************************************
#       Copyright (C) 2016 Javier Honrubia Gonzalez <jhonrubia6@alumno.uned.es>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from sage.matrix.constructor import matrix
from sage.misc.decorators import options, rename_keyword
from sage.misc.functional import is_odd
from sage.misc.lazy_import import lazy_import
from sage.plot.hyperbolic_polygon import HyperbolicPolygon
from sage.plot.plot import Graphics
from sage.rings.cc import CC
from sage.rings.integer import Integer

lazy_import("sage.functions.hyperbolic", "arccosh")
lazy_import("sage.functions.trig", ["sin", "cos", "cot"])
lazy_import("sage.symbolic.constants", ["pi", "e"])


class HyperbolicRegularPolygon(HyperbolicPolygon):
    r"""
    Primitive class for regular hyperbolic polygon type.

    See ``hyperbolic_regular_polygon?`` for information about plotting
    a hyperbolic regular polygon in the upper complex halfplane.

    INPUT:

    - ``sides`` -- number of sides of the polygon

    - ``i_angle`` -- interior angle of the polygon

    - ``center`` -- center point as a complex number of the polygon

    EXAMPLES:

    Note that constructions should use :func:`hyperbolic_regular_polygon`::

         sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
         sage: print(HyperbolicRegularPolygon(5,pi/2,I, {}))
         Hyperbolic regular polygon (sides=5, i_angle=1/2*pi, center=1.00000000000000*I)

    The code verifies is there exists a compact hyperbolic regular polygon
    with the given data, checking

    .. MATH::

        A(\mathcal{P}) = \pi(s-2) - s \cdot \alpha > 0,

    where `s` is ``sides`` and `\alpha` is ``i_angle``. This raises an error if
    the ``i_angle`` is less than the minimum to generate a compact polygon::

        sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
        sage: P = HyperbolicRegularPolygon(4, pi/2, I, {})
        Traceback (most recent call last):
        ...
        ValueError: there exists no hyperbolic regular compact polygon,
         for sides=4 the interior angle must be less than 1/2*pi

    It is an error to give a center outside the upper half plane in this model ::

         sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
         sage: P = HyperbolicRegularPolygon(4, pi/4, 1-I, {})
         Traceback (most recent call last):
         ...
         ValueError: center: 1.00000000000000 - 1.00000000000000*I is not
          a valid point in the upper half plane model of the hyperbolic plane

    TESTS::

         sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
         sage: P = HyperbolicRegularPolygon(4, -pi/4, I, {})
         Traceback (most recent call last):
         ...
         ValueError: interior angle -1/4*pi must be in (0, pi) interval

         sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
         sage: P=HyperbolicRegularPolygon(16, 3*pi/2, I, {})
         Traceback (most recent call last):
         ...
         ValueError: interior angle 3/2*pi must be in (0, pi) interval

         sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
         sage: P = HyperbolicRegularPolygon(2, pi/10, I, {})
         Traceback (most recent call last):
         ...
         ValueError: degenerated polygons (sides<=2) are not supported
    """
    def __init__(self, sides, i_angle, center, options):
        """
        Initialize HyperbolicRegularPolygon.

        EXAMPLES::

            sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
            sage: print(HyperbolicRegularPolygon(5,pi/2,I, {}))
            Hyperbolic regular polygon (sides=5, i_angle=1/2*pi, center=1.00000000000000*I)
        """
        self.center = CC(center)
        if self.center.imag() <= 0:
            raise ValueError("center: %s is not a valid point in the upper half plane model of the hyperbolic plane" % (self.center))
        if sides < 3:
            raise ValueError("degenerated polygons (sides<=2) are not supported")
        if i_angle <= 0 or i_angle >= pi:
            raise ValueError("interior angle %s must be in (0, pi) interval" % (i_angle))
        if pi*(sides-2) - sides*i_angle <= 0:
            raise ValueError("there exists no hyperbolic regular compact polygon,"
                             " for sides={} the interior angle must be less than {}".format(sides, pi * (sides-2) / sides))
        self.sides = sides
        self.i_angle = i_angle
        beta = 2 * pi / self.sides # compute the rotation angle to be used ahead
        alpha = self.i_angle / Integer(2)
        I = CC(0, 1)
        # compute using cosine theorem the radius of the circumscribed circle
        # using the triangle formed by the radius and the three known angles
        r = arccosh(cot(alpha) * (1 + cos(beta)) / sin(beta))

        # The first point will be always on the imaginary axis limited
        # to 8 digits for efficiency in the subsequent calculations.
        z_0 = [I*(e**r).n(digits=8)]

        # Compute the dilation isometry used to move the center
        # from I to the imaginary part of the given center.
        scale = self.center.imag()

        # Compute the parabolic isometry to move the center to the
        # real part of the given center.
        h_disp = self.center.real()

        d_z_k = [z_0[0]*scale + h_disp]  # d_k has the points for the polygon in the given center
        z_k = z_0                      # z_k has the Re(z)>0 vertices for the I centered polygon
        r_z_k = []                     # r_z_k has the Re(z)<0 vertices
        if is_odd(self.sides):
            vert = (self.sides - 1) // 2
        else:
            vert = self.sides // 2 - 1
        for k in range(vert):
            # Compute with 8 digits to accelerate calculations
            new_z_k = self._i_rotation(z_k[-1], beta).n(digits=8)
            z_k = z_k + [new_z_k]
            d_z_k = d_z_k + [new_z_k * scale + h_disp]
            r_z_k = [-(new_z_k).conjugate() * scale + h_disp] + r_z_k
        if is_odd(self.sides):
            HyperbolicPolygon.__init__(self, d_z_k + r_z_k, "UHP", options)
        else:
            z_opo = [I * (e**(-r)).n(digits=8) * scale + h_disp]
            HyperbolicPolygon.__init__(self, d_z_k + z_opo + r_z_k, "UHP", options)

    def _repr_(self):
        """
        String representation of HyperbolicRegularPolygon.

        TESTS::

            sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
            sage: HyperbolicRegularPolygon(5,pi/2,I, {})
            Hyperbolic regular polygon (sides=5, i_angle=1/2*pi, center=1.00000000000000*I)
        """
        return ("Hyperbolic regular polygon (sides={}, i_angle={}, center={})".format(self.sides, self.i_angle, self.center))

    def _i_rotation(self, z, alpha):
        r"""
        Return the resulting point after applying a hyperbolic
        rotation centered at `0 + i` and angle ``alpha`` to ``z``.

        INPUT:

        - ``z`` -- point in the upper complex halfplane to which
          apply the isometry

        - ``alpha`` -- angle of rotation (radians, counterclockwise)

        OUTPUT: rotated point in the upper complex halfplane

        TESTS::

            sage: from sage.plot.hyperbolic_regular_polygon import HyperbolicRegularPolygon
            sage: P = HyperbolicRegularPolygon(4, pi/4, 1+I, {})
            sage: P._i_rotation(2+I, pi/2)
            I - 2
        """
        _a = alpha / 2
        _c = cos(_a)
        _s = sin(_a)
        G = matrix([[_c, _s], [-_s, _c]])
        return (G[0][0] * z + G[0][1]) / (G[1][0] * z + G[1][1])


@rename_keyword(color='rgbcolor')
@options(alpha=1, fill=False, thickness=1, rgbcolor='blue', zorder=2,
         linestyle='solid')
def hyperbolic_regular_polygon(sides, i_angle, center=CC(0,1), **options):
    r"""
    Return a hyperbolic regular polygon in the upper half model of
    Hyperbolic plane given the number of sides, interior angle and
    possibly a center.

    Type ``?hyperbolic_regular_polygon`` to see all options.

    INPUT:

    - ``sides`` -- number of sides of the polygon

    - ``i_angle`` -- interior angle of the polygon

    - ``center`` -- (default: `i`) hyperbolic center point
      (complex number) of the polygon

    OPTIONS:

    - ``alpha`` -- (default: 1)

    - ``fill`` -- (default: ``False``)

    - ``thickness`` -- (default: 1)

    - ``rgbcolor`` -- (default: ``'blue'``)

    - ``linestyle`` -- (default: ``'solid'``) the style of the line,
      which can be one of the following:

      * ``'dashed'`` or ``'--'``
      * ``'dotted'`` or ``':'``
      * ``'solid'`` or ``'-'``
      * ``'dashdot'`` or ``'-.'``

    EXAMPLES:

    Show a hyperbolic regular polygon with 6 sides and square angles::

        sage: g = hyperbolic_regular_polygon(6, pi/2)
        sage: g.plot()
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

         g = hyperbolic_regular_polygon(6, pi/2)
         sphinx_plot(g.plot())

    With more options::

        sage: g = hyperbolic_regular_polygon(6, pi/2, center=3+2*I, fill=True, color='red')
        sage: g.plot()
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

         g = hyperbolic_regular_polygon(6, pi/2, center=3+2*I, fill=True, color='red')
         sphinx_plot(g.plot())

    The code verifies is there exists a hyperbolic regular polygon
    with the given data, checking

    .. MATH::

        A(\mathcal{P}) = \pi(s-2) - s \cdot \alpha > 0,

    where `s` is ``sides`` and `\alpha` is ``i_angle``. This raises an error if
    the ``i_angle`` is less than the minimum to generate a compact polygon::

        sage: hyperbolic_regular_polygon(4, pi/2)
        Traceback (most recent call last):
        ...
        ValueError: there exists no hyperbolic regular compact polygon,
         for sides=4 the interior angle must be less than 1/2*pi

    It is an error to give a center outside the upper half plane in
    this model::

        sage: from sage.plot.hyperbolic_regular_polygon import hyperbolic_regular_polygon
        sage: hyperbolic_regular_polygon(4, pi/4, 1-I)
        Traceback (most recent call last):
        ...
        ValueError: center: 1.00000000000000 - 1.00000000000000*I is not
         a valid point in the upper half plane model of the hyperbolic plane
    """
    g = Graphics()
    g._set_extra_kwds(g._extract_kwds_for_show(options))
    g.add_primitive(HyperbolicRegularPolygon(sides, i_angle, center, options))
    g.set_aspect_ratio(1)
    return g



# cython: wraparound=False, boundscheck=False
# sage.doctest: needs sage.graphs
r"""
Reflection groups: auxiliary Cython functions

This contains a few time-critical auxiliary cython functions for
finite complex or real reflection groups.
"""
# ****************************************************************************
#       Copyright (C) 2011-2016 Christian Stump <christian.stump at gmail.com>
#                     2016 Travis Scrimshaw <tscrimsh at umn.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.groups.perm_gps.permgroup_element cimport PermutationGroupElement
from collections import deque
from cysignals.memory cimport sig_malloc
from cpython.list cimport *

cdef class Iterator():
    """
    Iterator class for reflection groups.
    """
    cdef int n
    cdef int N  # number of reflections / positive roots
    cdef tuple S
    cdef str algorithm
    cdef bint tracking_words
    cdef list noncom
    cdef list order

    cdef list noncom_letters(self):
        """
        Return a list ``L`` of lists such that ...

        .. WARNING:: This is not used as it slows down the computation.

        EXAMPLES::

            sage: from sage.combinat.root_system.reflection_group_c import Iterator
            sage: W = ReflectionGroup(["B", 4])               # optional - gap3
            sage: I = Iterator(W, W.number_of_reflections())  # optional - gap3
            sage: TestSuite(I).run(skip='_test_pickling')     # optional - gap3
        """
        cdef tuple S = self.S
        cdef int n = len(S)
        cdef list noncom = []
        cdef list noncom_i
        for i in range(n):
            si = S[i]
            noncom_i = []
            for j in range(i+1, n):
                sj = S[j]
                if si._mul_(sj) == sj._mul_(si):
                    pass
                else:
                    noncom_i.append(j)
            noncom.append(noncom_i)
        noncom.append(list(range(n)))
        return noncom

    def __init__(self, W, int N, str algorithm='depth', bint tracking_words=True,
                 order=None):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.combinat.root_system.reflection_group_c import Iterator
            sage: W = ReflectionGroup(["B", 4])               # optional - gap3
            sage: I = Iterator(W, W.number_of_reflections())  # optional - gap3
            sage: TestSuite(I).run(skip='_test_pickling')     # optional - gap3
        """
        self.S = tuple(W.simple_reflections())
        self.n = len(W._index_set)
        self.N = N
        self.tracking_words = tracking_words

        if order is None:
            self.order = list(range(self.n))

        # "breadth" is 1.5x slower than "depth" since it uses
        # a deque with popleft instead of a list with pop
        if algorithm not in ["depth", "breadth", "parabolic"]:
            raise ValueError('the algorithm (="%s") must be either "depth", "breadth", or "parabolic"')
        self.algorithm = algorithm

        # self.noncom = self.noncom_letters()

    cdef list succ(self, PermutationGroupElement u, int first):
        cdef PermutationGroupElement si
        cdef int i
        cdef list successors = []
        cdef tuple S = self.S
        cdef int N = self.N
        # cdef list nc = self.noncom[first]

        for i in range(first):
            si = <PermutationGroupElement>(S[i])
            if self.test(u, si, i):
                successors.append((_new_mul_(si, u), i))
        for i in range(first+1, self.n):
            # for i in nc:
            if u.perm[i] < N:
                si = <PermutationGroupElement>(S[i])
                if self.test(u, si, i):
                    successors.append((_new_mul_(si, u), i))
        return successors

    cdef list succ_words(self, PermutationGroupElement u, list word, int first):
        cdef PermutationGroupElement u1, si
        cdef int i
        cdef list successors = []
        cdef list word_new
        cdef tuple S = self.S
        cdef int N = self.N

        for i in range(first):
            si = <PermutationGroupElement>(S[i])
            if self.test(u, si, i):
                u1 = <PermutationGroupElement>(_new_mul_(si, u))
                # try to use word+[i] and the reversed
                word_new = [i] + word
                u1._reduced_word = word_new
                successors.append((u1, word_new, i))
        for i in range(first+1, self.n):
            if u.perm[i] < N:
                si = <PermutationGroupElement>(S[i])
                if self.test(u, si, i):
                    u1 = <PermutationGroupElement>(_new_mul_(si, u))
                    word_new = [i] + word
                    u1._reduced_word = word_new
                    successors.append((u1, word_new, i))
        return successors

    cdef inline bint test(self, PermutationGroupElement u, PermutationGroupElement si, int i) noexcept:
        cdef int j
        cdef int N = self.N
        cdef int* siperm = si.perm
        cdef int* uperm = u.perm

        for j in range(i):
            if uperm[siperm[j]] >= N:
                return False
        return True

    def __iter__(self):
        """
        EXAMPLES::

            sage: # optional - gap3
            sage: from sage.combinat.root_system.reflection_group_c import Iterator
            sage: W = ReflectionGroup(["B", 4])
            sage: N = W.number_of_reflections()
            sage: I = Iterator(W, N)
            sage: len(list(I)) == W.cardinality()
            True

            sage: # optional - gap3
            sage: I = Iterator(W, N, "breadth", False)
            sage: len(list(I)) == W.cardinality()
            True
            sage: I = Iterator(W, N, "parabolic")
            sage: len(list(I)) == W.cardinality()
            True
        """
        # the breadth search iterator is ~2x slower as it
        # uses a deque with popleft
        if self.algorithm == "depth":
            if self.tracking_words:
                return self.iter_words_depth()
            else:
                return self.iter_depth()
        elif self.algorithm == "breadth":
            if self.tracking_words:
                return self.iter_words_breadth()
            else:
                return self.iter_breadth()
        elif self.algorithm == "parabolic":
            return self.iter_parabolic()

    def iter_depth(self):
        """
        Iterate over ``self`` using depth-first-search.

        EXAMPLES::

            sage: from sage.combinat.root_system.reflection_group_c import Iterator
            sage: W = CoxeterGroup(['B',2], implementation='permutation')
            sage: I = Iterator(W, W.number_of_reflections())
            sage: list(I.iter_depth())
            [(),
             (1,3)(2,6)(5,7),
             (1,5)(2,4)(6,8),
             (1,3,5,7)(2,8,6,4),
             (1,7)(3,5)(4,8),
             (1,7,5,3)(2,4,6,8),
             (2,8)(3,7)(4,6),
             (1,5)(2,6)(3,7)(4,8)]
        """
        cdef list cur = [(self.S[0].parent().one(), -1)]
        cdef PermutationGroupElement u
        cdef int first
        cdef list L = []

        while True:
            if not cur:
                if not L:
                    return
                cur = L.pop()
                continue

            u, first = cur.pop()
            yield u
            L.append(self.succ(u, first))

    def iter_words_depth(self):
        """
        Iterate over ``self`` using depth-first-search and setting
        the reduced word.

        EXAMPLES::

            sage: from sage.combinat.root_system.reflection_group_c import Iterator
            sage: W = CoxeterGroup(['B',2], implementation='permutation')
            sage: I = Iterator(W, W.number_of_reflections())
            sage: for w in I.iter_words_depth(): w._reduced_word
            []
            [1]
            [0]
            [1, 0]
            [0, 1, 0]
            [0, 1]
            [1, 0, 1]
            [0, 1, 0, 1]
        """
        cdef list cur, word

        cdef PermutationGroupElement u
        cdef int first
        cdef list L = []

        one = self.S[0].parent().one()
        one._reduced_word = []
        cur = [(one, list(), -1)]

        while True:
            if not cur:
                if not L:
                    return
                cur = L.pop()
                continue

            u, word, first = cur.pop()
            yield u
            L.append(self.succ_words(u, word, first))

    def iter_breadth(self):
        """
        Iterate over ``self`` using breadth-first-search.

        EXAMPLES::

            sage: from sage.combinat.root_system.reflection_group_c import Iterator
            sage: W = CoxeterGroup(['B',2], implementation='permutation')
            sage: I = Iterator(W, W.number_of_reflections())
            sage: list(I.iter_breadth())
            [(),
             (1,3)(2,6)(5,7),
             (1,5)(2,4)(6,8),
             (1,7,5,3)(2,4,6,8),
             (1,3,5,7)(2,8,6,4),
             (2,8)(3,7)(4,6),
             (1,7)(3,5)(4,8),
             (1,5)(2,6)(3,7)(4,8)]
        """
        cdef list cur = [(self.S[0].parent().one(), -1)]
        cdef PermutationGroupElement u
        cdef int first
        L = deque()

        while True:
            if not cur:
                if not L:
                    return
                cur = L.popleft()
                continue

            u, first = cur.pop()
            yield u
            L.append(self.succ(u, first))

    def iter_words_breadth(self):
        """
        Iterate over ``self`` using breadth-first-search and setting
        the reduced word.

        EXAMPLES::

            sage: from sage.combinat.root_system.reflection_group_c import Iterator
            sage: W = CoxeterGroup(['B',2], implementation='permutation')
            sage: I = Iterator(W, W.number_of_reflections())
            sage: for w in I.iter_words_breadth(): w._reduced_word
            []
            [1]
            [0]
            [0, 1]
            [1, 0]
            [1, 0, 1]
            [0, 1, 0]
            [0, 1, 0, 1]
        """
        cdef list cur, word
        cdef PermutationGroupElement u
        cdef int first
        L = deque()

        one = self.S[0].parent().one()
        one._reduced_word = []
        cur = [(one, list(), -1)]

        while True:
            if not cur:
                if not L:
                    return
                cur = L.popleft()
                continue

            u, word, first = cur.pop()
            yield u
            L.append(self.succ_words(u, word, first))

    def iter_parabolic(self):
        r"""
        This algorithm is an alternative to the one in *chevie* and about
        20% faster. It yields indeed all elements in the group rather than
        applying a given function.

        The output order is not deterministic.

        EXAMPLES::

            sage: from sage.combinat.root_system.reflection_group_c import Iterator
            sage: W = CoxeterGroup(['B',2], implementation='permutation')
            sage: I = Iterator(W, W.number_of_reflections())
            sage: sorted(I.iter_parabolic())
            [(),
             (2,8)(3,7)(4,6),
             (1,3)(2,6)(5,7),
             (1,3,5,7)(2,8,6,4),
             (1,5)(2,4)(6,8),
             (1,5)(2,6)(3,7)(4,8),
             (1,7)(3,5)(4,8),
             (1,7,5,3)(2,4,6,8)]
        """
        cdef int i
        cdef list coset_reps
        W = self.S[0].parent()

        cdef list elts = [W.one()]

        for i in range(1, self.n):
            coset_reps = reduced_coset_representatives(W, self.order[:i],
                                                       self.order[:i-1], True)
            elts = [_new_mul_(<PermutationGroupElement>w, <PermutationGroupElement>v)
                    for w in elts for v in coset_reps]
        # the list ``elts`` now contains all prods of red coset reps

        coset_reps = reduced_coset_representatives(W, self.order,
                                                   self.order[:len(self.order)-1], True)

        for w in elts:
            for v in coset_reps:
                yield _new_mul_(<PermutationGroupElement>w, <PermutationGroupElement>v)


def iterator_tracking_words(W):
    r"""
    Return an iterator through the elements of ``self`` together
    with the words in the simple generators.

    The iterator is a breadth first search through the graph of the
    elements of the group with generators.

    EXAMPLES::

        sage: from sage.combinat.root_system.reflection_group_c import iterator_tracking_words
        sage: W = ReflectionGroup(4)                        # optional - gap3
        sage: for w in iterator_tracking_words(W): w        # optional - gap3
        ((), [])
        ((1,3,9)(2,4,7)(5,10,18)(6,11,16)(8,12,19)(13,15,20)(14,17,21)(22,23,24), [0])
        ((1,5,13)(2,6,10)(3,7,14)(4,8,15)(9,16,22)(11,12,17)(18,19,23)(20,21,24), [1])
        ((1,9,3)(2,7,4)(5,18,10)(6,16,11)(8,19,12)(13,20,15)(14,21,17)(22,24,23), [0, 0])
        ((1,7,6,12,23,20)(2,8,17,24,9,5)(3,16,10,19,15,21)(4,14,11,22,18,13), [0, 1])
        ((1,10,4,12,21,22)(2,11,19,24,13,3)(5,15,7,17,16,23)(6,18,8,20,14,9), [1, 0])
        ((1,13,5)(2,10,6)(3,14,7)(4,15,8)(9,22,16)(11,17,12)(18,23,19)(20,24,21), [1, 1])
        ((1,16,12,15)(2,14,24,18)(3,5,19,17)(4,6,22,20)(7,8,23,9)(10,13,21,11), [0, 0, 1])
        ((1,2,12,24)(3,6,19,20)(4,17,22,5)(7,11,23,13)(8,21,9,10)(14,16,18,15), [0, 1, 0])
        ((1,14,12,18)(2,15,24,16)(3,22,19,4)(5,6,17,20)(7,10,23,21)(8,11,9,13), [0, 1, 1])
        ((1,18,12,14)(2,16,24,15)(3,4,19,22)(5,20,17,6)(7,21,23,10)(8,13,9,11), [1, 0, 0])
        ((1,15,12,16)(2,18,24,14)(3,17,19,5)(4,20,22,6)(7,9,23,8)(10,11,21,13), [1, 1, 0])
        ((1,6,23)(2,17,9)(3,10,15)(4,11,18)(5,8,24)(7,12,20)(13,14,22)(16,19,21), [0, 0, 1, 0])
        ((1,22,21,12,4,10)(2,3,13,24,19,11)(5,23,16,17,7,15)(6,9,14,20,8,18), [0, 0, 1, 1])
        ((1,4,21)(2,19,13)(3,11,24)(5,7,16)(6,8,14)(9,18,20)(10,12,22)(15,17,23), [0, 1, 0, 0])
        ((1,17,13,12,5,11)(2,20,10,24,6,21)(3,23,14,19,7,18)(4,9,15,22,8,16), [0, 1, 1, 0])
        ((1,19,9,12,3,8)(2,22,7,24,4,23)(5,21,18,17,10,14)(6,13,16,20,11,15), [1, 0, 0, 1])
        ((1,20,23,12,6,7)(2,5,9,24,17,8)(3,21,15,19,10,16)(4,13,18,22,11,14), [1, 1, 0, 0])
        ((1,11,5,12,13,17)(2,21,6,24,10,20)(3,18,7,19,14,23)(4,16,8,22,15,9), [0, 0, 1, 0, 0])
        ((1,23,6)(2,9,17)(3,15,10)(4,18,11)(5,24,8)(7,20,12)(13,22,14)(16,21,19), [0, 0, 1, 1, 0])
        ((1,8,3,12,9,19)(2,23,4,24,7,22)(5,14,10,17,18,21)(6,15,11,20,16,13), [0, 1, 0, 0, 1])
        ((1,21,4)(2,13,19)(3,24,11)(5,16,7)(6,14,8)(9,20,18)(10,22,12)(15,23,17), [0, 1, 1, 0, 0])
        ((1,12)(2,24)(3,19)(4,22)(5,17)(6,20)(7,23)(8,9)(10,21)(11,13)(14,18)(15,16), [0, 0, 1, 0, 0, 1])
        ((1,24,12,2)(3,20,19,6)(4,5,22,17)(7,13,23,11)(8,10,9,21)(14,15,18,16), [0, 0, 1, 1, 0, 0])
    """
    cdef tuple S = tuple(W.simple_reflections())
    cdef list index_list = list(range(len(S)))

    cdef list level_set_cur = [(W.one(), [])]
    cdef set level_set_old = {W.one()}
    cdef list word
    cdef PermutationGroupElement x, y

    while level_set_cur:
        level_set_new = []
        for x, word in level_set_cur:
            yield x, word
            for i in index_list:
                y = _new_mul_(x, <PermutationGroupElement>(S[i]))
                if y not in level_set_old:
                    level_set_old.add(y)
                    level_set_new.append((y, word+[i]))
        level_set_cur = level_set_new


cdef inline bint has_left_descent(PermutationGroupElement w, int i, int N) noexcept:
    return w.perm[i] >= N

cdef int first_descent(PermutationGroupElement w, int n, int N, bint left) noexcept:
    cdef int i
    if not left:
        w = ~w
    for i in range(n):
        if has_left_descent(w, i, N):
            return i
    return -1

cdef int first_descent_in_parabolic(PermutationGroupElement w, list parabolic,
                                    int N, bint left) noexcept:
    cdef int i
    if not left:
        w = ~w
    # this loop over a list might be slow
    for i in parabolic:
        if has_left_descent(w, i, N):
            return i
    return -1


cpdef PermutationGroupElement reduce_in_coset(PermutationGroupElement w, tuple S,
                                              list parabolic, int N, bint right):
    r"""
    Return the minimal length coset representative of ``w`` of the parabolic
    subgroup indexed by ``parabolic`` (with indices `\{0, \ldots, n\}`).

    EXAMPLES::

        sage: from sage.combinat.root_system.reflection_group_c import reduce_in_coset
        sage: W = CoxeterGroup(['B',3], implementation='permutation')
        sage: N = W.number_of_reflections()
        sage: s = W.simple_reflections()
        sage: w = s[2] * s[1] * s[3]
        sage: reduce_in_coset(w, tuple(s), [], N, True).reduced_word()
        [2, 1, 3]
        sage: reduce_in_coset(w, tuple(s), [], N, False).reduced_word()
        [2, 1, 3]
        sage: reduce_in_coset(w, tuple(s), [0], N, True).reduced_word()
        [2, 1, 3]
        sage: reduce_in_coset(w, tuple(s), [0], N, False).reduced_word()
        [2, 3]
        sage: reduce_in_coset(w, tuple(s), [0,2], N, True).reduced_word()
        [2, 1, 3]
        sage: reduce_in_coset(w, tuple(s), [0,2], N, False).reduced_word()
        [2]
    """
    cdef int i
    cdef PermutationGroupElement si

    if right:
        while True:
            i = first_descent_in_parabolic(w, parabolic, N, True)
            if i == -1:
                return w
            si = <PermutationGroupElement>(S[i])
            w = _new_mul_(si, w)
    else:
        while True:
            i = first_descent_in_parabolic(w, parabolic, N, False)
            if i == -1:
                return w
            si = <PermutationGroupElement>(S[i])
            w = _new_mul_(w, si)

cdef list reduced_coset_representatives(W, list parabolic_big, list parabolic_small,
                                        bint right):
    cdef tuple S = tuple(W.simple_reflections())
    cdef int N = W.number_of_reflections()
    cdef set totest = set([W.one()])
    cdef set res = set(totest)
    cdef set new

    while totest:
        new = set()
        for w in totest:
            new.update([reduce_in_coset(_new_mul_(w, <PermutationGroupElement>(S[i])),
                                        S, parabolic_small, N, right)
                        for i in parabolic_big])
        res.update(totest)
        totest = new.difference(res)  # [w for w in new if w not in res]
    return list(res)

cdef parabolic_recursive(PermutationGroupElement x, list v, f):
    if not v:
        f(x)
    else:
        for y in v[0]:
            parabolic_recursive(_new_mul_(x, <PermutationGroupElement>y), v[1:], f)


def parabolic_iteration_application(W, f):
    r"""
    This is the word-for-word translation of the algorithm in chevie.

    .. NOTE::

        It keeps all products of elements of the reduced coset
        representatives in memory.

    INPUT:

    - ``W`` -- a real reflection group
    - ``f`` -- a function with one argument: an element of ``W``

    EXAMPLES::

        sage: W = CoxeterGroup(['E',6], implementation='permutation')
        sage: from sage.combinat.root_system.reflection_group_c import parabolic_iteration_application
        sage: lst = []
        sage: def f(x):
        ....:     lst.append(x)
        sage: parabolic_iteration_application(W, f)
        sage: len(lst) == W.cardinality()
        True
    """
    cdef int i
    cdef list coset_reps = [reduced_coset_representatives(W, list(range(i+1)),
                                                          list(range(i)), True)
                            for i in range(W.rank())]

    parabolic_recursive(W.one(), coset_reps, f)


cpdef list reduced_word_c(W, PermutationGroupElement w):
    r"""
    Compute a reduced word for the element ``w`` in the
    reflection group ``W`` in the positions ``range(n)``.

    EXAMPLES::

        sage: from sage.combinat.root_system.reflection_group_c import reduced_word_c
        sage: W = ReflectionGroup(['B',2])                  # optional - gap3
        sage: [ reduced_word_c(W,w) for w in W ]            # optional - gap3
        [[], [1], [0], [0, 1], [1, 0], [1, 0, 1], [0, 1, 0], [0, 1, 0, 1]]
    """
    cdef tuple S = tuple(W.simple_reflections())
    cdef int n = len(S)
    cdef int N = W.number_of_reflections()
    cdef int fdes = 0
    cdef list word = []

    while True:
        fdes = first_descent(w, n, N, True)
        if fdes == -1:
            break
        w = _new_mul_(<PermutationGroupElement>(S[fdes]), w)
        word.append(fdes)
    return word

cdef PermutationGroupElement _new_mul_(PermutationGroupElement left, PermutationGroupElement right):
    """
    Multiply two :class:`PermutationGroupElement` directly without the
    coercion framework.
    """
    cdef type t = type(left)
    cdef PermutationGroupElement prod = t.__new__(t)
    cdef int n = left.n
    cdef int sizeofint = sizeof(int)
    cdef int n_sizeofint = sizeofint * n

#    if HAS_DICTIONARY(left):
#        prod.__class__ = left.__class__
    prod._parent = left._parent
    prod.n = n
    if n_sizeofint <= sizeof(prod.perm_buf):
        prod.perm = prod.perm_buf
    else:
        prod.perm = <int *> sig_malloc(n_sizeofint)

    cdef int i
    for i in range(n):
        prod.perm[i] = right.perm[left.perm[i]]

    return prod
